import logging
from pathlib import Path

from cg.io.controller import ReadFile
from cg.utils.files import get_file_in_directory, get_files_in_directory_with_pattern

LOG = logging.getLogger(__name__)


class ValidateFileTransferService:
    """Service to validate file transfers via a manifest file."""

    def validate_file_transfer(
        self, manifest_file: Path, source_dir: Path, manifest_file_format: str
    ):
        """Validate all files listed in the manifest are present in the directory tree."""
        files_in_manifest: list[str] = self._get_files_in_manifest(
            manifest_file=manifest_file, manifest_file_format=manifest_file_format
        )
        if not self._are_all_files_present(
            files_to_validate=files_in_manifest, source_dir=source_dir
        ):
            raise FileNotFoundError(
                f"Not all files listed in the manifest file {manifest_file} are present in the directory tree {source_dir}"
            )
        LOG.debug("File transfer validated")
        return

    @staticmethod
    def _get_manifest_file_content(manifest_file: Path, manifest_file_format: str) -> list[str]:
        """Get the content of the manifest file."""
        file_reader = ReadFile()
        return file_reader.get_content_from_file(
            file_format=manifest_file_format, file_path=manifest_file
        )

    @staticmethod
    def _is_valid_path(line: str):
        return "/" in line

    def _extract_file_names_from_manifest(self, manifest_content: any) -> list[str]:
        """
        Extract the file paths from the manifest content.
        A file path is expected to contain at least one '/' character.
        """
        file_names: list[str] = []
        for line in manifest_content:
            if self._is_valid_path(line):
                formatted_line: str = line.rstrip()
                file_names.append(Path(formatted_line).name)
        return file_names

    @staticmethod
    def _is_file_in_directory_tree(file_name: str, source_dir: Path) -> bool:
        """Check if a file is present in the directory tree."""
        try:
            if get_file_in_directory(directory=source_dir, file_name=file_name):
                return True
        except FileNotFoundError:
            return False

    @staticmethod
    def _get_manifest_file_paths(source_dir: Path, pattern: str) -> list[Path]:
        return get_files_in_directory_with_pattern(directory=source_dir, pattern=pattern)

    def _get_files_in_manifest(self, manifest_file: Path, manifest_file_format: str) -> list[str]:
        """Get the files listed in the manifest file."""
        manifest_content: list[str] = self._get_manifest_file_content(
            manifest_file=manifest_file, manifest_file_format=manifest_file_format
        )
        return self._extract_file_names_from_manifest(manifest_content)

    def _are_all_files_present(self, files_to_validate: list[str], source_dir: Path) -> bool:
        """Check if all files are present in the directory tree."""
        for file_name in files_to_validate:
            if not self._is_file_in_directory_tree(file_name=file_name, source_dir=source_dir):
                return False
        return True



# This one needs run_dir, out_dir, basemask and sample_sheet

DEMULTIPLEX_COMMAND: str = """
log "dragen --bcl-conversion-only true \
--bcl-input-directory {run_dir} \
--output-directory {demux_dir} \
--force
touch {demux_completed_file}"

dragen --bcl-conversion-only true \
--bcl-input-directory {run_dir} \
--output-directory {demux_dir} \
--force
touch {demux_completed_file}
log "Dragen BCL Convert finished!"
"""

# This needs flow cell id, email. logfile
DEMULTIPLEX_ERROR = """
mail -s 'ERROR demultiplexing of {flow_cell_id}' {email} < '{logfile}'
if [[ -e {demux_dir} ]]
then
    rm -r {demux_dir}
fi
if [[ -e {demux_started} ]]
then
    rm -r {demux_started}
fi
"""



"""Test for analysis"""

import logging
from datetime import datetime

import mock
import pytest

from cg.constants import GenePanelMasterList, Priority, SequencingRunDataAvailability
from cg.constants.archiving import ArchiveLocations
from cg.constants.constants import ControlOptions
from cg.constants.priority import SlurmQos, TrailblazerPriority
from cg.constants.sequencing import Sequencers
from cg.exc import AnalysisNotReadyError
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.meta.workflow.utils.utils import are_all_samples_control, MAP_TO_TRAILBLAZER_PRIORITY
from cg.models.fastq import FastqFileMeta
from cg.store.models import Case, Sample, IlluminaSequencingRun
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.mark.parametrize(
    "priority,expected_slurm_qos",
    [
        (Priority.clinical_trials, SlurmQos.NORMAL),
        (Priority.research, SlurmQos.LOW),
        (Priority.standard, SlurmQos.NORMAL),
        (Priority.priority, SlurmQos.HIGH),
        (Priority.express, SlurmQos.EXPRESS),
    ],
)
def test_get_slurm_qos_for_case(
    case_id: str,
    priority,
    expected_slurm_qos,
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
):
    """Test get Quality of service (SLURM QOS) from the case priority"""

    # GIVEN a store with a case with a specific priority
    mip_analysis_api.status_db = analysis_store
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    case.priority = priority

    # WHEN getting the SLURM QOS for the priority
    slurm_qos: SlurmQos = mip_analysis_api.get_slurm_qos_for_case(case_id=case_id)

    # THEN the expected slurm QOS should be returned
    assert slurm_qos is expected_slurm_qos


@pytest.mark.parametrize(
    "priority,expected_trailblazer_priority",
    [
        (Priority.clinical_trials, TrailblazerPriority.NORMAL),
        (Priority.research, TrailblazerPriority.LOW),
        (Priority.standard, TrailblazerPriority.NORMAL),
        (Priority.priority, TrailblazerPriority.HIGH),
        (Priority.express, TrailblazerPriority.EXPRESS),
    ],
)
def test_get_trailblazer_priority(
    case_id: str,
    priority,
    expected_trailblazer_priority,
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
):
    """Test get Trailblazer priority from the case priority"""

    # GIVEN a store with a case with a specific priority
    mip_analysis_api.status_db = analysis_store
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    case.priority = priority

    # WHEN getting the trailblazer priority for the case
    trailblazer_priority: TrailblazerPriority = mip_analysis_api.get_trailblazer_priority(
        case_id=case_id
    )

    # THEN the expected trailblazer priority should be returned
    assert trailblazer_priority is expected_trailblazer_priority


def test_gene_panels_not_part_of_master_list(customer_id: str):
    """Test get only broad non-specific gene panels and custom gene panel list if a supplied gene panels is not part of master list."""
    # GIVEN a customer who is a collaborator on the master list

    # GIVEN a gene panel NOT included in the gene panel master list
    default_panels_not_included: list[str] = ["PANEL_NOT_IN_GENE_PANEL_MASTER_LIST"]

    # WHEN converting the gene panels between the default and the gene panel master list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id=customer_id, default_panels=set(default_panels_not_included)
    )

    # THEN the list of gene panels returned is the custom panel and broad non-specific panels
    assert set(
        list_of_gene_panels_used
    ) == GenePanelMasterList.get_non_specific_gene_panels().union(set(default_panels_not_included))


def test_gene_panels_customer_collaborator_and_panel_part_of_master_list(customer_id: str):
    """Test get gene panels when customer is collaborator for gene panel master list, and supplied gene panels are a subset of the masters list."""

    # GIVEN a a gene panel included in the gene panel master list
    default_panels_included: list[str] = [GenePanelMasterList.get_panel_names()[0]]

    # GIVEN a master list

    # WHEN converting the gene panels between the default and the gene panel master list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id=customer_id, default_panels=set(default_panels_included)
    )

    # THEN the list of gene panels used should return all gene panels in the master list
    assert set(list_of_gene_panels_used) == set(GenePanelMasterList.get_panel_names())


def test_gene_panels_customer_is_collaborator_and_panel_not_part_of_master_list(customer_id: str):
    """Test get gene panels when customer is collaborator for gene panel master list, but supplied gene panels are not a subset of the masters list."""

    # GIVEN a case that has a customer collaborator for the gene panel master list

    # GIVEN a gene panel not in the master list
    default_panels: list[str] = ["A_PANEL_NOT_IN_MASTER_LIST"]

    # WHEN converting the gene panels between the default and the gene panel master list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id=customer_id, default_panels=set(default_panels)
    )

    # THEN the list of gene panels returned the custom panel and broad non-specific panels
    assert set(
        list_of_gene_panels_used
    ) == GenePanelMasterList.get_non_specific_gene_panels().union(set(default_panels))


def test_gene_panels_customer_not_collaborator_for_gene_master_list():
    """Test get gene panels when customer is not collaborator for the gene panel master list."""
    # GIVEN a a gene panel included in the gene panel master list
    default_panels: list[str] = [GenePanelMasterList.get_panel_names()[0]]

    # GIVEN a case that has a customer, who is not a collaborator for the gene panel master list

    # WHEN converting the gene panels between the default and the gene panel master list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id="CUSTOMER_NOT_COLLABORATOR_FOR_MASTER_LIST", default_panels=set(default_panels)
    )

    # THEN the list of gene panels returned the custom panel and broad non-specific panels
    assert set(
        list_of_gene_panels_used
    ) == GenePanelMasterList.get_non_specific_gene_panels().union(set(default_panels))


def test_is_illumina_run_check_applicable(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store
):
    """Tests that a check for flow cells being present on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # GIVEN that no samples are down-sampled nor external
    for sample in case.samples:
        assert not sample.from_sample

    # WHEN checking if a flow cell check is applicable
    # THEN the method should return True
    assert mip_analysis_api._is_illumina_run_check_applicable(case_id=case.internal_id)


def test_is_flow_cell_check_not_applicable_when_external(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store
):
    """Tests that a check for flow cells being present on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # WHEN marking all of its samples as external
    for sample in case.samples:
        sample.application_version.application.is_external = True

    # WHEN checking if a flow cell check is applicable
    # THEN the method should return False
    assert not mip_analysis_api._is_illumina_run_check_applicable(case_id=case.internal_id)


def test_is_illumina_run_check_not_applicable_when_down_sampled(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that a check for Illumina runs being on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = store_with_illumina_sequencing_data_on_disk.get_case_by_internal_id(
        selected_novaseq_x_case_ids[0]
    )

    # WHEN marking all of its samples as down sampled from TestSample
    for sample in case.samples:
        sample.from_sample = "TestSample"

    # WHEN checking if an Illumina run check is applicable
    # THEN the method should return False
    assert not mip_analysis_api._is_illumina_run_check_applicable(case_id=case.internal_id)


def test_ensure_illumina_run_on_disk_check_not_applicable(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
    caplog,
):
    """Tests that ensure_illumina_run_on_disk does not perform any action
    when is_illumina_runs_check_applicable returns false."""
    caplog.set_level(logging.INFO)
    # GIVEN a case
    case: Case = store_with_illumina_sequencing_data_on_disk.get_case_by_internal_id(
        selected_novaseq_x_case_ids[0]
    )

    # WHEN _is_flow_cell_check_available returns False
    with mock.patch.object(
        AnalysisAPI,
        "_is_illumina_run_check_applicable",
        return_value=False,
    ):
        mip_analysis_api.ensure_illumina_run_on_disk(case.internal_id)

    # THEN a warning should be logged
    assert (
        "Illumina run check is not applicable - the case is either down sampled or external."
        in caplog.text
    )


def test_ensure_illumina_runs_on_disk_does_not_request_runs(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    helpers: StoreHelpers,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that ensure_illumina_run_on_disk does not perform any action
    when is_Illumina_runs_check_applicable returns True and all runs are ON_DISK already."""

    # GIVEN a case

    # WHEN _is_illumina_run_check_available returns True and the attached flow cell is ON_DISK
    with mock.patch.object(
        AnalysisAPI,
        "_is_illumina_run_check_applicable",
        return_value=True,
    ), mock.patch.object(
        Store, "request_sequencing_runs_for_case", return_value=None
    ) as request_checker:
        mip_analysis_api.ensure_illumina_run_on_disk(selected_novaseq_x_case_ids[0])

    # THEN runs should not be requested for the case
    assert request_checker.call_count == 0


def test_ensure_illumina_run_on_disk_does_request_run(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    helpers: StoreHelpers,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that ensure_illumina_run_on_disk requests a removed flow cell
    when is_flow_cell_check_applicable returns True.."""

    # GIVEN a case with a REMOVED Illumina run
    sequencing_runs: list[IlluminaSequencingRun] = (
        store_with_illumina_sequencing_data_on_disk.get_illumina_sequencing_runs_by_case(
            selected_novaseq_x_case_ids[0]
        )
    )
    for sequencing_run in sequencing_runs:
        sequencing_run.data_availability = SequencingRunDataAvailability.REMOVED

    # WHEN _is_flow_cell_check_available returns True
    with mock.patch.object(
        AnalysisAPI,
        "_is_illumina_run_check_applicable",
        return_value=True,
    ):
        mip_analysis_api.ensure_illumina_run_on_disk(selected_novaseq_x_case_ids[0])

    # THEN the flow cell's status should be set to REQUESTED for the case
    modified_runs: list[IlluminaSequencingRun] = (
        store_with_illumina_sequencing_data_on_disk.get_illumina_sequencing_runs_by_case(
            selected_novaseq_x_case_ids[0]
        )
    )
    for run in modified_runs:
        assert run.data_availability == SequencingRunDataAvailability.REQUESTED


def test_is_case_ready_for_analysis_true(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that is_case_ready_for_analysis returns true for a case whose illumina runs are all ON_DISK and whose
    files need no decompression nor are being decompressed currently."""
    # GIVEN a case and a flow cell with status ON_DISK

    # GIVEN that no decompression is needed nor running
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=False):
        # WHEN running is_case_ready_for_analysis

        # THEN the result should be true
        assert mip_analysis_api.is_raw_data_ready_for_analysis(selected_novaseq_x_case_ids[0])


def test_is_case_ready_for_analysis_decompression_needed(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that is_case_ready_for_analysis returns false for a case whose sequencing runs are all ON_DISK but whose
    files need decompression."""

    # GIVEN a case and an Illumina run

    # GIVEN that some files need to be decompressed
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=True
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=False):
        # WHEN running is_case_ready_for_analysis
        # THEN the result should be false
        assert not mip_analysis_api.is_raw_data_ready_for_analysis(selected_novaseq_x_case_ids[0])


def test_is_case_ready_for_analysis_decompression_running(
    mip_hk_store,
    selected_novaseq_x_case_ids: list[str],
    mip_analysis_api: MipDNAAnalysisAPI,
):
    """Tests that is_case_ready_for_analysis returns false for a case whose Illumina run are all ON_DISK but whose
    files are being decompressed currently."""

    # GIVEN a case and a Illumina sequencing run

    # GIVEN that some files are being decompressed
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=True):
        # WHEN running is_case_ready_for_analysis
        # THEN the result should be false
        assert not mip_analysis_api.is_raw_data_ready_for_analysis(selected_novaseq_x_case_ids[0])


@pytest.mark.parametrize(
    "is_spring_decompression_needed, is_spring_decompression_running",
    [(False, False), (True, False), (False, True), (True, True)],
)
def test_prepare_fastq_files_success(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
    helpers,
    is_spring_decompression_needed,
    is_spring_decompression_running,
):
    """Tests that no error is thrown when running prepare_fastq_files when its Illumina runs are all ON_DISK,
    and no spring decompression is needed nor is running."""

    # GIVEN a case with an Illumina run ON_DISK

    # GIVEN that no decompression is or running and adding the files to Housekeeper goes well
    with mock.patch.object(
        PrepareFastqAPI,
        "is_spring_decompression_needed",
        return_value=is_spring_decompression_needed,
    ), mock.patch.object(MipAnalysisAPI, "decompress_case", return_value=None), mock.patch.object(
        PrepareFastqAPI,
        "is_spring_decompression_running",
        return_value=is_spring_decompression_running,
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ):
        # WHEN running prepare_fastq_files
        if is_spring_decompression_running or is_spring_decompression_needed:
            with pytest.raises(AnalysisNotReadyError):
                mip_analysis_api.prepare_fastq_files(
                    case_id=selected_novaseq_x_case_ids[0], dry_run=False
                )
        else:
            mip_analysis_api.prepare_fastq_files(
                case_id=selected_novaseq_x_case_ids[0], dry_run=False
            )


def test_prepare_fastq_files_request_miria(
    mip_analysis_api: MipDNAAnalysisAPI,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that samples' input files are requested via Miria for a Clinical customer, if files are archived."""

    # GIVEN a case belonging to a non-PDC customer with at least one archived spring file
    case: Case = mip_analysis_api.status_db.get_case_by_internal_id(selected_novaseq_x_case_ids[0])
    case.customer.data_archive_location = ArchiveLocations.KAROLINSKA_BUCKET

    # GIVEN that at least one file is archived and not retrieved
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_running", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ), mock.patch.object(
        SpringArchiveAPI, "retrieve_spring_files_for_case"
    ), mock.patch.object(
        AnalysisAPI, "is_raw_data_ready_for_analysis", return_value=False
    ) as request_submitter:
        with pytest.raises(AnalysisNotReadyError):
            # WHEN running prepare_fastq_files

            # THEN an AnalysisNotReadyError should be thrown
            mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)

    # THEN retrieve_samples should have been invoked
    assert request_submitter.call_count == 1


def test_prepare_fastq_files_does_not_request_miria(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    helpers: StoreHelpers,
    selected_novaseq_x_case_ids: list[str],
):
    """Tests that samples' input files are not requested via Miria for a Clinical customer, if no files are archived."""

    # GIVEN a case belonging to a non-PDC customer

    # GIVEN that no files have entries in the Archive table
    case: Case = store_with_illumina_sequencing_data_on_disk.get_case_by_internal_id(
        selected_novaseq_x_case_ids[0]
    )
    case.customer.data_archive_location = ArchiveLocations.KAROLINSKA_BUCKET

    # GIVEN that all Illumina runs have status on disk
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_running", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ):
        # WHEN running prepare_fastq_files

        # THEN no error should be raised
        mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)


def test_are_all_spring_files_present_true_when_all_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    non_archived_spring_file,
    case_id: str,
    father_sample_id: str,
    helpers: StoreHelpers,
):
    """Tests that are_all_spring_files_present returns True when no files for a case are archived."""

    # GIVEN that the only Spring file belonging to a case has no Archive entry

    # WHEN checking if all spring files are present

    # THEN the result should be True
    assert mip_analysis_api.are_all_spring_files_present(case_id)


def test_are_all_spring_files_present_false_when_none_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    archived_spring_file,
    case_id: str,
    sample_id: str,
):
    """Tests that are_all_spring_files_present returns False when all files for a case are archived and not retrieved."""

    # GIVEN that the only Spring file belonging to a case has an Archive entry which has retrieved_at not set

    # WHEN checking if all spring files are present

    # THEN the result should be False
    assert not mip_analysis_api.are_all_spring_files_present(case_id)


def test_ensure_files_are_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    archived_spring_file,
    case_id: str,
    sample_id: str,
):
    """Tests that archived files are requested via Miria after the flow cell check has been performed."""

    # GIVEN that the only Spring file belonging to a case has an Archive entry which has retrieved_at not set

    # WHEN ensuring that all files are present

    with mock.patch.object(AnalysisAPI, "ensure_illumina_run_on_disk"), mock.patch.object(
        SpringArchiveAPI,
        "retrieve_spring_files_for_case",
    ) as request_submitter:
        mip_analysis_api.ensure_files_are_present(case_id)

    # THEN the files should have been requested
    request_submitter.assert_called()


@pytest.mark.parametrize(
    "data_availability, result",
    [(SequencingRunDataAvailability.REMOVED, True), (SequencingRunDataAvailability.ON_DISK, False)],
)
def test_does_any_spring_file_need_to_be_retrieved_flow_cell_status(
    mip_analysis_api: MipDNAAnalysisAPI,
    store_with_illumina_sequencing_data_on_disk: Store,
    selected_novaseq_x_case_ids: list[str],
    helpers: StoreHelpers,
    data_availability: str,
    result: bool,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns True if one of
    the Illumina run has data availability 'removed' and False if the status is instead 'ondisk'."""

    # GIVEN a case with an illumina with provided data availability
    sequencing_runs: list[IlluminaSequencingRun] = (
        store_with_illumina_sequencing_data_on_disk.get_illumina_sequencing_runs_by_case(
            selected_novaseq_x_case_ids[0]
        )
    )
    for run in sequencing_runs:
        run.data_availability = data_availability

    # WHEN checking if any files need to be retrieved

    # THEN the outcome should be False if the Illumina run data availability is on disk but not otherwise
    assert (
        mip_analysis_api.does_any_file_need_to_be_retrieved(selected_novaseq_x_case_ids[0])
        == result
    )


def test_does_any_spring_file_need_to_be_retrieved_archived_file(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    case_id: str,
    archived_spring_file,
    non_archived_spring_file,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns true if no flow cell is removed but some
    files need to be retrieved via Miria."""

    assert mip_analysis_api.does_any_file_need_to_be_retrieved(case_id)


def test_does_any_spring_file_need_to_be_retrieved_files_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    case_id: str,
    non_archived_spring_file,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns true if no flow cell is removed and no
    files need to be retrieved via Miria."""

    assert mip_analysis_api.does_any_file_need_to_be_retrieved(case_id)


def test_link_fastq_files_for_sample(
    analysis_store: Store,
    caplog,
    mip_analysis_api: MipDNAAnalysisAPI,
    fastq_file_meta_raw: dict,
    mocker,
):
    caplog.set_level(logging.INFO)
    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # GIVEN a sample
    sample: Sample = case.links[0].sample

    with mocker.patch.object(
        AnalysisAPI,
        "gather_file_metadata_for_sample",
        return_value=[FastqFileMeta.model_validate(fastq_file_meta_raw)],
    ):
        # WHEN parsing header
        mip_analysis_api.link_fastq_files_for_sample(case=case, sample=sample)

        # THEN broadcast linking of files
        assert "Linking: " in caplog.text


def test_are_all_samples_control(analysis_store: Store, case_id: str) -> None:
    """Tests that are_all_samples_control returns True if all samples in a case are controls."""

    # GIVEN a case with all samples being positive controls
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    for sample in case.samples:
        sample.control = ControlOptions.POSITIVE

    # WHEN checking if all samples are controls
    # THEN the result should be True
    assert are_all_samples_control(case)


@pytest.mark.parametrize(
    "sample_controls, expected_qos",
    [
        (
            [ControlOptions.POSITIVE, ControlOptions.POSITIVE, ControlOptions.POSITIVE],
            SlurmQos.EXPRESS,
        ),
        (
            [ControlOptions.POSITIVE, ControlOptions.NEGATIVE, ControlOptions.EMPTY],
            SlurmQos.NORMAL,
        ),
    ],
    ids=["all_controls", "not_all_controls"],
)
def test_case_with_controls_get_correct_slurmqos(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    case_id: str,
    sample_controls: list[str],
    expected_qos: str,
) -> None:
    """Tests that get_slurm_qos_for_case returns the correct SLURM QOS for a case with control samples."""

    # GIVEN a case with the specified sample control types
    mip_analysis_api.status_db = analysis_store
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    for index, control in enumerate(sample_controls):
        case.samples[index].control = control

    # WHEN getting the SLURM QOS for the case
    qos: SlurmQos = mip_analysis_api.get_slurm_qos_for_case(case_id)

    # THEN the result should match the expected QOS
    assert qos == expected_qos


@pytest.mark.parametrize(
    "priority, expected_trailblazer_priority",
    [
        (Priority.clinical_trials, TrailblazerPriority.NORMAL),
        (Priority.research, TrailblazerPriority.LOW),
        (Priority.standard, TrailblazerPriority.NORMAL),
        (Priority.priority, TrailblazerPriority.HIGH),
        (Priority.express, TrailblazerPriority.EXPRESS),
    ],
)
def test_get_trailblazer_priority(
    case_id: str,
    priority: Priority,
    expected_trailblazer_priority: TrailblazerPriority,
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
):
    """Test get Trailblazer priority from the case priority"""

    # GIVEN a store with a case with a specific priority
    mip_analysis_api.status_db = analysis_store
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    case.priority = priority

    # WHEN getting the trailblazer priority for the priority
    trailblazer_priority: TrailblazerPriority = mip_analysis_api.get_trailblazer_priority(
        case_id=case_id
    )

    # THEN the expected trailblazer priority should be returned
    assert trailblazer_priority is expected_trailblazer_priority



import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_formatter.destination.base_service import BaseDeliveryFormatter
from cg.services.deliver_files.tag_fetcher.bam_service import (
    BamDeliveryTagsFetcher,
)
from cg.services.deliver_files.tag_fetcher.fohm_upload_service import FOHMUploadTagsFetcher
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.raw_data_service import (
    RawDataDeliveryFileFetcher,
)
from cg.services.deliver_files.file_formatter.files.case_service import (
    CaseFileFormatter,
)
from cg.services.deliver_files.file_formatter.files.sample_service import (
    SampleFileFormatter,
    FileManager,
)
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)
from cg.store.store import Store


@pytest.fixture
def raw_data_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def raw_data_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def bam_data_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = BamDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def bam_data_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = BamDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def fohm_data_delivery_service(
    delivery_fohm_upload_housekeeper_api: HousekeeperAPI,
    delivery_store_mutant: Store,
) -> RawDataAndAnalysisDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = FOHMUploadTagsFetcher()
    return RawDataAndAnalysisDeliveryFileFetcher(
        hk_api=delivery_fohm_upload_housekeeper_api,
        status_db=delivery_store_mutant,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def analysis_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_balsamic: Store,
) -> AnalysisDeliveryFileFetcher:
    """Fixture to get an instance of FetchAnalysisDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return AnalysisDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_balsamic,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def analysis_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_balsamic: Store,
) -> AnalysisDeliveryFileFetcher:
    """Fixture to get an instance of FetchAnalysisDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return AnalysisDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_balsamic,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def generic_delivery_file_formatter() -> BaseDeliveryFormatter:
    """Fixture to get an instance of GenericDeliveryFileFormatter."""
    return BaseDeliveryFormatter(
        sample_file_formatter=SampleFileFormatter(
            file_manager=FileManager(), path_name_formatter=NestedStructurePathFormatter()
        ),
        case_file_formatter=CaseFileFormatter(
            file_manager=FileManager(),
            path_name_formatter=NestedStructurePathFormatter(),
        ),
    )



"""Loqusdb upload output fixtures."""

import pytest

LOQUSDB_OUTPUT = (
    b"2018-11-29 08:41:38 130-229-8-20-dhcp.local "
    b"mongo_adapter.client[77135] INFO Connecting to "
    b"uri:mongodb://None@localhost:27017\n"
    b"2018-11-29 08:41:38 130-229-8-20-dhcp.local "
    b"mongo_adapter.client[77135] INFO Connection "
    b"established\n2018-11-29 08:41:38 130-229-8-20-dhcp.local "
    b"mongo_adapter.adapter[77135] INFO"
    b" Use database loqusdb\n2018-11-29 08:41:38 130-229-8-20-dhcp.local "
    b"loqusdb.utils.vcf[77135]"
    b" INFO Check if vcf is on correct format...\n"
    b"2018-11-29 08:41:38 130-229-8-20-dhcp.local"
    b" loqusdb.utils.vcf[77135] INFO Vcf file "
    b"/Users/username/Projects/loqusdb/tests/fixtures"
    b"/test.vcf.gz looks fine\n2018-11-29 "
    b"08:41:38 130-229-8-20-dhcp.local loqusdb.utils.vcf[77135]"
    b" INFO Nr of variants in vcf: 15\n2018-11-29 "
    b"08:41:38 130-229-8-20-dhcp.local loqusdb.utils."
    b"vcf[77135] INFO Type of variants in vcf: snv\n"
    b"Inserting variants\n2018-11-29 08:41:38 130-22"
    b"9-8-20-dhcp.local loqusdb.utils.load[77135] "
    b"INFO Inserted 15 variants of type snv\n2018-11-2"
    b"9 08:41:38 130-229-8-20-dhcp.local loqusdb.commands.load[77135] "
    b"INFO Nr variants inserted: "
    b"15\n2018-11-29 08:41:38 130-229-8-20-dhcp.local "
    b"loqusdb.commands.load[77135] INFO Time to "
    b"insert variants: 0:00:00.012648\n2018-11-29 "
    b"08:41:38 130-229-8-20-dhcp.local loqusdb.plugins."
    b"mongo.adapter[77135] INFO All indexes exists\n"
)

LOQUSDB_CASE_OUTPUT = (
    b'[{"_id": "1234", "case_id": "yellowhog", '
    b'"vcf_path": "test.vcf.gz", "vcf_sv_path": null, "nr_variants": 15, '
    b'"nr_sv_variants": null, "profile_path": "test.vcf.gz", "individuals": '
    b'[{"ind_id": "proband", "name": "proband", "case_id": "recessive_trio", '
    b'"ind_index": 2, "sex": 1, "profile": ["CC", "TT", "TT", "CC", "TT", '
    b'"AA", "CC", "GG", "GG", "GG", "TT", "GG", "CC", "AA", "GG", "GG", "GG", '
    b'"AA", "CC", "CC", "CC", "GG", "CC", "CC", "TT", "CC", "GG", "GG", "TT", '
    b'"AA", "AA", "GG", "CC", "AA", "AA", "AA", "AA", "GG", "GG", "TT", "GG", '
    b'"AA", "TT", "AA", "GG", "GG", "CC", "AA", "TT", "TT"]}, {"ind_id": '
    b'"mother", "name": "mother", "case_id": "recessive_trio", "ind_index": 1, '
    b'"sex": 2, "profile": ["CC", "TT", "TT", "CC", "TT", "AA", "CC", "GG", '
    b'"GG", "GG", "TT", "GG", "CC", "AA", "GG", "GG", "GG", "AA", "CC", "CC", '
    b'"CC", "GG", "CC", "CC", "TT", "CC", "GG", "GG", "TT", "AA", "AA", "GG", '
    b'"CC", "AA", "AA", "AA", "AA", "GG", "GG", "TT", "GG", "AA", "TT", "AA", '
    b'"GG", "GG", "CC", "AA", "TT", "TT"]}, {"ind_id": "father", "name": '
    b'"father", "case_id": "recessive_trio", "ind_index": 0, "sex": 1, '
    b'"profile": ["CC", "TT", "TT", "CC", "TT", "AA", "CC", "GG", "GG", '
    b'"GG", "TT", "GG", "CC", "AA", "GG", "GG", "GG", "AA", "CC", "CC", '
    b'"CC", "GG", "CC", "CC", "TT", "CC", "GG", "GG", "TT", "AA", "AA", '
    b'"GG", "CC", "AA", "AA", "AA", "AA", "GG", "GG", "TT", "GG", "AA", '
    b'"TT", "AA", "GG", "GG", "CC", "AA", "TT", "TT"]}], "sv_individuals": '
    b'[], "_inds": {"proband": {"ind_id": "proband", "name": "proband", '
    b'"case_id": "recessive_trio", "ind_index": 2, "sex": 1}, "mother": '
    b'{"ind_id": "mother", "name": "mother", "case_id": "recessive_trio", '
    b'"ind_index": 1, "sex": 2}, "father": {"ind_id": "father", "name": '
    b'"father", "case_id": "recessive_trio", "ind_index": 0, "sex": 1}}, '
    b'"_sv_inds": {}}]\n'
)

LOQUSDB_DUPLICATE_OUTPUT = (
    b'{"ind_id": "proband", "name": "proband", "case_id": "recessive_trio_test", '
    b'"ind_index": 4, "sex": 1, "profile": ["TT", "CC", "CC", "CC", "TT", "AA",'
    b' "CC", "GG", "GG", "GG", "TT", "GG", "CC", "AA", "GG", "GG", "GG", "AA", '
    b'"CC", "CC", "CC", "GG", "CC", "CC", "TT", "CC", "AA", "GA", "TC", "AG", "AC", '
    b'"GT", "CC", "AC", "AG", "AG", "AA", "GA", "AA", "CC", "GG", "AA", "TT", "AA", '
    b'"GG", "GG", "CC", "AA", "TT", "TT"]}'
)

LOQUSDB_DELETE_STDERR = b"""2022-09-22 12:30:07 username loqusdb.commands.cli[20689] INFO Running loqusdb version 2.6.9
2022-09-22 12:30:07 username mongo_adapter.client[20689] INFO Connecting to uri:mongodb://None@localhost:27017
2022-09-22 12:30:07 username mongo_adapter.client[20689] INFO Connection established
2022-09-22 12:30:07 username mongo_adapter.adapter[20689] INFO Use database loqusdb
2022-09-22 12:30:07 username loqusdb.plugins.mongo.case[20689] INFO Removing case yellowhog from database
2022-09-22 12:30:07 username loqusdb.utils.delete[20689] INFO deleting variants
2022-09-22 12:30:07 username loqusdb.utils.delete[20689] INFO Start deleting chromosome 1"""

LOQUSDB_DELETE_NONEXISTING_STDERR = b"""2022-09-22 11:40:04 username loqusdb.commands.cli[19944] INFO Running loqusdb version 2.6.9
2022-09-22 11:40:04 username mongo_adapter.client[19944] INFO Connecting to uri:mongodb://None@localhost:27017
2022-09-22 11:40:04 username mongo_adapter.client[19944] INFO Connection established
2022-09-22 11:40:04 username mongo_adapter.adapter[19944] INFO Use database loqusdb
2022-09-22 11:40:04 username loqusdb.commands.delete[19944] WARNING Case yellowhog does not exist in database"""


@pytest.fixture
def loqusdb_load_stderr() -> bytes:
    """Return Loqusdb stderr for a successful load."""
    return LOQUSDB_OUTPUT


@pytest.fixture
def loqusdb_case_output() -> bytes:
    """Return Loqusdb output for a 'loqusdb cases -c <case_id> --to-json' command."""
    return LOQUSDB_CASE_OUTPUT


@pytest.fixture
def loqusdb_duplicate_output() -> bytes:
    """Return Loqusdb output for a 'loqusdb profile --check-vcf' command."""
    return LOQUSDB_DUPLICATE_OUTPUT


@pytest.fixture
def loqusdb_delete_stderr() -> bytes:
    """Return Loqusdb STDERR for a successful delete."""
    return LOQUSDB_DELETE_STDERR


@pytest.fixture
def loqusdb_delete_non_existing_stderr() -> bytes:
    """Return Loqusdb delete STDERR for non existing case."""
    return LOQUSDB_DELETE_NONEXISTING_STDERR


@pytest.fixture
def number_of_loaded_variants() -> int:
    """Return number of loaded variants."""
    return 15



"""Code for uploading to Scout via CLI."""

import logging
from pathlib import Path

import rich_click as click
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.cli.upload.utils import suggest_cases_to_upload
from cg.constants import Workflow
from cg.constants.cli_options import DRY_RUN
from cg.constants.constants import FileFormat
from cg.constants.scout import ScoutCustomCaseReportTags
from cg.io.controller import WriteStream
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.meta.workflow.tomte import TomteAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("scout")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def upload_to_scout(context, re_upload: bool, print_console: bool, case_id: str):
    """Upload variants from analysis to Scout."""
    status_db: Store = context.obj.status_db

    LOG.info("----------------- SCOUT -----------------------")

    if not case_id:
        suggest_cases_to_upload(status_db=status_db)
        return

    context.invoke(
        create_scout_load_config,
        case_id=case_id,
        print_console=print_console,
        re_upload=re_upload,
    )
    if not print_console:
        context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command(name="create-scout-load-config")
@click.argument("case-id")
@click.option("-p", "--print", "print_console", is_flag=True, help="Only print config")
@click.option("-r", "--re-upload", is_flag=True, help="Overwrite existing load configs")
@click.pass_obj
def create_scout_load_config(context: CGConfig, case_id: str, print_console: bool, re_upload: bool):
    """Create a load config for a case in scout and add it to housekeeper"""

    status_db: Store = context.status_db

    LOG.info("Fetching family object")
    case_obj: Case = status_db.get_case_by_internal_id(internal_id=case_id)

    if not case_obj.analyses:
        LOG.warning(f"Could not find analyses for {case_id}")
        raise click.Abort

    context.meta_apis["upload_api"]: UploadAPI = get_upload_api(cg_config=context, case=case_obj)

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api

    LOG.info("----------------- CREATE CONFIG -----------------------")
    LOG.info("Create load config")
    try:
        scout_load_config: ScoutLoadConfig = scout_upload_api.generate_config(case_obj.analyses[0])
    except SyntaxError as error:
        LOG.warning(repr(error))
        raise click.Abort from error
    LOG.info(f"Found load config: {scout_load_config}")

    root_dir: str = context.meta_apis["upload_api"].analysis_api.root
    LOG.info(f"Set root dir to {root_dir}")
    file_path: Path = Path(root_dir, case_id, "scout_load.yaml")
    if print_console:
        click.echo(
            WriteStream.write_stream_from_content(
                content=scout_load_config.dict(exclude_none=True), file_format=FileFormat.YAML
            )
        )
        LOG.info(f"Would save file to {file_path}")
        return

    if file_path.exists():
        LOG.warning(f"Scout load config {file_path} already exists")
        if re_upload:
            LOG.info("Deleting old load config")
            file_path.unlink()
        else:
            LOG.info(
                "You might remove the file and try again, consider that you might also have it in housekeeper"
            )
            raise click.Abort
    LOG.info("Saving config file to disc")
    scout_upload_api.save_config_file(upload_config=scout_load_config, file_path=file_path)

    try:
        LOG.info("Add config file to housekeeper")
        scout_upload_api.add_scout_config_to_hk(
            config_file_path=file_path, case_id=case_id, delete=re_upload
        )
    except FileExistsError as error:
        LOG.warning(f"{error}, consider removing the file from housekeeper and try again")
        raise click.Abort from error


@click.command(name="upload-case-to-scout")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_case_to_scout(context: CGConfig, re_upload: bool, dry_run: bool, case_id: str):
    """Upload variants and case from analysis to Scout."""

    LOG.info("----------------- UPLOAD -----------------------")

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    scout_api: ScoutAPI = context.scout_api

    tag_name: str = UploadScoutAPI.get_load_config_tag()
    version: Version = housekeeper_api.last_version(bundle=case_id)
    scout_config_file: File | None = housekeeper_api.get_latest_file_from_version(
        version=version, tags={tag_name}
    )

    if scout_config_file is None:
        raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

    LOG.info(f"Uploading case {case_id} to scout")

    if not dry_run:
        scout_api.upload(scout_load_config=scout_config_file.full_path, force=re_upload)

    LOG.info(f"Uploaded to scout using load config {scout_config_file.full_path}")
    LOG.info("Case loaded successfully to Scout")


@click.command(name="validate-case-samples-are-rna")
@click.argument("case_id")
@click.pass_obj
def validate_case_samples_are_rna(context: CGConfig, case_id: str) -> None:
    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    are_all_samples_rna: bool = scout_upload_api.analysis_api.are_case_samples_rna(case_id)
    if not are_all_samples_rna:
        LOG.error(f"{case_id} has non-RNA samples - aborting.")
        raise click.Abort


@click.command(name="rna-to-scout")
@DRY_RUN
@click.option("-r", "--research", is_flag=True, help="Upload research report instead of clinical")
@click.argument("case_id")
@click.pass_context
def upload_rna_to_scout(context, case_id: str, dry_run: bool, research: bool) -> None:
    """Upload an RNA case's gene fusion report and junction splice files for all samples connected via subject ID."""
    LOG.info("----------------- UPLOAD RNA TO SCOUT -----------------------")

    context.invoke(validate_case_samples_are_rna, case_id=case_id)
    context.invoke(upload_rna_alignment_file_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(upload_multiqc_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(
        upload_rna_fusion_report_to_scout, case_id=case_id, dry_run=dry_run, research=research
    )
    context.invoke(upload_rna_junctions_to_scout, case_id=case_id, dry_run=dry_run)


@click.command(name="tomte-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_context
def upload_tomte_to_scout(
    context,
    case_id: str,
    dry_run: bool,
) -> None:
    """Upload an Tomte RNA case's junction splice files and omics files for all samples connected via subject ID."""
    LOG.info("----------------- UPLOAD RNA TO SCOUT -----------------------")

    context.invoke(validate_case_samples_are_rna, case_id=case_id)
    context.invoke(upload_rna_delivery_report_to_scout, case_id=case_id)
    context.invoke(upload_rna_alignment_file_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(upload_multiqc_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(upload_rna_omics_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(upload_rna_junctions_to_scout, case_id=case_id, dry_run=dry_run)


@click.command(name="rna-alignment-file-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_alignment_file_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload RNA alignment file for a case to Scout."""
    LOG.info("----------------- UPLOAD RNA ALIGNMENT FILE TO SCOUT -----------------------")
    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_rna_alignment_file(case_id=case_id, dry_run=dry_run)


@click.command(name="rna-fusion-report-to-scout")
@DRY_RUN
@click.option("--research", is_flag=True)
@click.argument("case_id")
@click.pass_obj
def upload_rna_fusion_report_to_scout(
    context: CGConfig, dry_run: bool, case_id: str, research: bool
) -> None:
    """Upload fusion report file for a case to Scout."""

    LOG.info("----------------- UPLOAD RNA FUSION REPORT TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_fusion_report_to_scout(
        dry_run=dry_run, research=research, case_id=case_id
    )


@click.command(name="rna-delivery-report-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_delivery_report_to_scout(context: CGConfig, dry_run: bool, case_id: str) -> None:
    """Upload rna delivery report file for a case to Scout."""

    LOG.info("----------------- UPLOAD RNA DELIVERY REPORT TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_rna_delivery_report_to_scout(dry_run=dry_run, case_id=case_id)


@click.command(name="rna-junctions-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_junctions_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload RNA junctions splice files to Scout."""
    LOG.info("----------------- UPLOAD RNA JUNCTIONS TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api

    scout_upload_api.upload_rna_junctions_to_scout(dry_run=dry_run, case_id=case_id)


@click.command(name="rna-omics-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_omics_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload RNA omics files to Scout."""
    LOG.info("----------------- UPLOAD RNA OMICS TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_rna_omics_to_scout(dry_run=dry_run, case_id=case_id)


@click.command(name="multiqc-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_multiqc_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload multiqc report to Scout."""
    LOG.info("----------------- UPLOAD MULTIQC TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    status_db: Store = context.status_db
    case: Case = status_db.get_case_by_internal_id(internal_id=case_id)
    scout_report_type, multiqc_report = scout_upload_api.get_multiqc_html_report(
        case_id=case_id, workflow=case.data_analysis
    )
    if scout_report_type == ScoutCustomCaseReportTags.MULTIQC_RNA:
        scout_upload_api.upload_rna_report_to_dna_case_in_scout(
            dry_run=dry_run,
            rna_case_id=case_id,
            report_type=scout_report_type,
            report_file=multiqc_report,
        )
    else:
        scout_upload_api.upload_report_to_scout(
            dry_run=dry_run,
            case_id=case_id,
            report_type=scout_report_type,
            report_file=multiqc_report,
        )


def get_upload_api(case: Case, cg_config: CGConfig) -> UploadAPI:
    """Return the upload API based on the data analysis type"""

    analysis_apis: dict[Workflow, UploadAPI] = {
        Workflow.BALSAMIC: BalsamicAnalysisAPI,
        Workflow.BALSAMIC_UMI: BalsamicUmiAnalysisAPI,
        Workflow.MIP_RNA: MipRNAAnalysisAPI,
        Workflow.MIP_DNA: MipDNAAnalysisAPI,
        Workflow.RAREDISEASE: RarediseaseAnalysisAPI,
        Workflow.RNAFUSION: RnafusionAnalysisAPI,
        Workflow.TOMTE: TomteAnalysisAPI,
    }

    return UploadAPI(
        config=cg_config, analysis_api=analysis_apis.get(case.data_analysis)(cg_config)
    )



"""Tests for the delivery report API."""

from pathlib import Path
from typing import Any

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture

from cg.constants import Workflow
from cg.exc import DeliveryReportError
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.models.analysis import AnalysisModel
from cg.models.delivery_report.metadata import SampleMetadataModel
from cg.models.delivery_report.report import (
    ReportModel,
    CustomerModel,
    CaseModel,
    DataAnalysisModel,
)
from cg.models.delivery_report.sample import (
    SampleModel,
    ApplicationModel,
    MethodsModel,
    TimestampModel,
)
from cg.store.models import Case, Analysis, Sample


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_delivery_report_html(request: FixtureRequest, workflow: Workflow):
    """Test rendering of the delivery report for different workflows."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case object
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # WHEN generating the delivery report HTML
    delivery_report_html: str = delivery_report_api.get_delivery_report_html(
        case_id=case_id, analysis_date=case.analyses[0].started_at, force=False
    )

    # THEN it should generate a valid HTML string
    assert delivery_report_html
    assert "<!DOCTYPE html>" in delivery_report_html


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def write_delivery_report_file(request: FixtureRequest, workflow: Workflow, tmp_path: Path):
    """Test writing of the delivery report for different workflows."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case object
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    delivery_report_file: Path = delivery_report_api.write_delivery_report_file(
        case_id=case.internal_id,
        directory=tmp_path,
        analysis_date=case.analyses[0].started_at,
        force=False,
    )

    # THEN check if the HTML delivery report has been generated and written
    assert delivery_report_file.is_file()


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_render_delivery_report(request: FixtureRequest, workflow: Workflow):
    """Test delivery report HTML rendering."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case object
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN some report data
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id=case_id, analysis_date=case.analyses[0].started_at
    )

    # WHEN rendering the report
    rendered_report: str = delivery_report_api.render_delivery_report(report_data.model_dump())

    # THEN validate rendered report
    assert rendered_report
    assert "<!DOCTYPE html>" in rendered_report


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_report_data(request: FixtureRequest, workflow: Workflow):
    """Test extraction of the data necessary to render the delivery report."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # WHEN extracting the report data
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id=case_id, analysis_date=case.analyses[0].started_at
    )

    # THEN the report model should have been populated and the data validated
    assert report_data
    assert isinstance(report_data, ReportModel)


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_validate_report_data(request: FixtureRequest, workflow: Workflow):
    """Test retrieval and validation of the delivery report data."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a report data model
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id=case_id, analysis_date=case.analyses[0].started_at
    )

    # WHEN validating the delivery report data
    report_data_validated: ReportModel = delivery_report_api.validate_report_data(
        case_id=case_id, report_data=report_data, force=False
    )

    # THEN the delivery report data should have been validated, with all the required fields present
    assert report_data_validated
    assert isinstance(report_data_validated, ReportModel)


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_validate_report_data_empty_optional_fields(
    request: FixtureRequest, workflow: Workflow, caplog: LogCaptureFixture
):
    """Test delivery report validations with allowed empty report fields."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a delivery report data model
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id, case.analyses[0].started_at
    )

    # GIVEN empty optional delivery report fields
    report_data.version = None
    report_data.customer.id = None
    report_data.case.samples[0].methods.library_prep = None

    # WHEN validating the delivery report data
    report_data: ReportModel = delivery_report_api.validate_report_data(
        case_id=case_id, report_data=report_data, force=False
    )

    # THEN the validation should not fail
    assert report_data

    # THEN the missing values should be logged
    assert "Empty report fields:" in caplog.text
    assert "version" in caplog.text
    assert "id" in caplog.text
    assert "library_prep" in caplog.text


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_validate_report_data_empty_required_fields(
    request: FixtureRequest, workflow: Workflow, caplog: LogCaptureFixture
):
    """Tests the validations of empty required delivery report fields."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a delivery report data model
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id, case.analyses[0].started_at
    )

    # GIVEN empty required delivery report fields
    report_data.accredited = None
    report_data.case.samples[0].metadata.million_read_pairs = None
    report_data.case.samples[0].metadata.duplicates = None

    # THEN a DeliveryReportError should be raised
    with pytest.raises(DeliveryReportError):
        delivery_report_api.validate_report_data(
            case_id=case_id, report_data=report_data, force=False
        )

    # THEN the required missing values should be logged
    assert "Missing data:" in caplog.text
    assert "accredited" in caplog.text
    assert "million_read_pairs" in caplog.text
    assert "duplicates" in caplog.text


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_validate_report_data_empty_required_fields_force(
    request: FixtureRequest, workflow: Workflow, caplog: LogCaptureFixture
):
    """Tests the validations of empty required delivery report fields with a force flag."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a delivery report data model
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id, case.analyses[0].started_at
    )

    # GIVEN empty required delivery report fields
    report_data.accredited = None
    report_data.case.samples[0].metadata.million_read_pairs = None
    report_data.case.samples[0].metadata.duplicates = None

    # THEN a DeliveryReportError should not be raised
    report_data_validated: ReportModel = delivery_report_api.validate_report_data(
        case_id=case_id, report_data=report_data, force=True
    )

    # THEN the delivery report data should have been validated despite missing required fields
    assert report_data_validated
    assert isinstance(report_data_validated, ReportModel)


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_validate_report_data_external_sample(request: FixtureRequest, workflow: Workflow):
    """Tests report data validation for a case with an external sample."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a delivery report data model
    report_data: ReportModel = delivery_report_api.get_report_data(
        case_id, case.analyses[0].started_at
    )

    # GIVEN a case with an external sample
    report_data.case.samples[0].timestamps.received_at = None
    report_data.case.samples[0].application.external = True

    # WHEN validating the delivery report fields
    report_data: ReportModel = delivery_report_api.validate_report_data(
        case_id=case_id, report_data=report_data, force=False
    )

    # THEN the delivery report data should have been successfully validated
    assert report_data


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_customer_data(request: FixtureRequest, workflow: Workflow):
    """Checks that the retrieved customer data is the expected one."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving the customers data
    customer_data: CustomerModel = delivery_report_api.get_customer_data(case)

    # THEN the customer model should have been successfully populated
    assert customer_data
    assert customer_data.name
    assert customer_data.id
    assert customer_data.invoice_address


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_report_version(request: FixtureRequest, workflow: Workflow):
    """Test report version extraction from an analysis object."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN an analysis object
    analysis: Analysis = case.analyses[0]

    # WHEN retrieving the delivery report version
    delivery_report_version: int = delivery_report_api.get_report_version(analysis)

    # THEN the first version should be returned
    assert delivery_report_version == 1


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_case_data(request: FixtureRequest, workflow: Workflow):
    """Test report version extraction from an analysis object."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN workflow specific analysis metadata
    analysis_metadata: AnalysisModel = delivery_report_api.analysis_api.get_latest_metadata(case_id)

    # WHEN retrieving case data
    case_data: CaseModel = delivery_report_api.get_case_data(
        case=case, analysis=case.analyses[0], analysis_metadata=analysis_metadata
    )

    # THEN the case delivery report model should have been populated
    assert case_data
    assert case_data.name
    assert case_data.data_analysis
    assert case_data.samples
    assert case_data.applications


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_samples_data(request: FixtureRequest, workflow: Workflow):
    """Test retrieval of sample data."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN workflow specific analysis metadata
    analysis_metadata: AnalysisModel = delivery_report_api.analysis_api.get_latest_metadata(case_id)

    # WHEN extracting the samples of a specific case
    samples_data: SampleModel = delivery_report_api.get_samples_data(
        case=case, analysis_metadata=analysis_metadata
    )[0]

    # THEN assert that the retrieved sample model has been correctly build
    assert samples_data
    assert samples_data.name
    assert samples_data.id
    assert samples_data.ticket
    assert samples_data.status
    assert samples_data.sex
    assert samples_data.source
    assert samples_data.tumour
    assert samples_data.application
    assert samples_data.methods
    assert samples_data.metadata
    assert samples_data.timestamps


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_sample_application(request: FixtureRequest, workflow: Workflow):
    """Test sample application data extraction."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a sample
    sample: Sample = case.samples[0]

    # GIVEN a LIMS sample
    lims_sample: dict[str, Any] = delivery_report_api.analysis_api.lims_api.sample(
        sample.internal_id
    )

    # WHEN retrieving application data from Statusdb
    application_data: ApplicationModel = delivery_report_api.get_sample_application(
        sample=sample, lims_sample=lims_sample
    )

    # THEN verify that the application data corresponds to what is expected
    assert application_data
    assert application_data.tag
    assert application_data.version
    assert application_data.prep_category
    assert application_data.description
    assert application_data.limitations
    assert application_data.accredited


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_unique_applications(request: FixtureRequest, workflow: Workflow):
    """Test unique applications filtering."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN workflow specific analysis metadata
    analysis_metadata: AnalysisModel = delivery_report_api.analysis_api.get_latest_metadata(case_id)

    # GIVEN a list of delivery report sample models
    samples: list[SampleModel] = delivery_report_api.get_samples_data(
        case=case, analysis_metadata=analysis_metadata
    )

    # WHEN filtering unique applications
    unique_applications: list[ApplicationModel] = delivery_report_api.get_unique_applications(
        samples
    )

    # THEN check that the same sample applications were filtered out
    assert len(unique_applications) == 1


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_sample_methods_data(request: FixtureRequest, workflow: Workflow, sample_id: str):
    """Test sample methods retrieval from LIMS."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a sample ID

    # WHEN extracting the methods metadata
    sample_methods: MethodsModel = delivery_report_api.get_sample_methods_data(sample_id)

    # THEN check sample methods data have been correctly retrieved
    assert sample_methods
    assert sample_methods.library_prep
    assert sample_methods.sequencing


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_case_analysis_data(request: FixtureRequest, workflow: Workflow):
    """Test data analysis retrieval."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving a case analysis
    case_analysis_data: DataAnalysisModel = delivery_report_api.get_case_analysis_data(
        case, case.analyses[0]
    )

    # THEN the case analysis data model has been populated
    assert case_analysis_data
    assert case_analysis_data.workflow
    assert case_analysis_data.panels
    assert case_analysis_data.scout_files


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_sample_timestamp_data(request: FixtureRequest, workflow: Workflow):
    """Test that the sample timestamp information is correctly retrieved from StatusDB."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a sample
    sample: Sample = case.samples[0]

    # WHEN extracting the timestamp data associated to a specific sample
    sample_timestamp_data: TimestampModel = delivery_report_api.get_sample_timestamp_data(sample)

    # THEN check if the dates are correctly retrieved
    assert sample_timestamp_data
    assert sample_timestamp_data.ordered_at
    assert sample_timestamp_data.received_at
    assert sample_timestamp_data.prepared_at
    assert sample_timestamp_data.reads_updated_at


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_get_sample_metadata(request: FixtureRequest, workflow: Workflow):
    """Test sample metadata extraction."""

    # GIVEN a delivery report API
    delivery_report_api: DeliveryReportAPI = request.getfixturevalue(
        f"{workflow}_delivery_report_api"
    )

    # GIVEN a case
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    case: Case = delivery_report_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a sample
    sample: Sample = case.samples[0]

    # GIVEN workflow specific analysis metadata
    analysis_metadata: AnalysisModel = delivery_report_api.analysis_api.get_latest_metadata(case_id)

    # WHEN retrieving the sample metadata
    sample_metadata: SampleMetadataModel = delivery_report_api.get_sample_metadata(
        case=case, sample=sample, analysis_metadata=analysis_metadata
    )

    # THEN check that the sample metadata has been correctly retrieved
    assert sample_metadata



from datetime import datetime

from cg.constants import DataDelivery, SexOptions, Workflow
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.services.orders.submitters.order_submitter import StoreOrderService
from cg.store.exc import EntryNotFoundError
from cg.store.models import Case, CaseSample, Customer, Order, Sample
from cg.store.store import Store


class StoreMicrobialFastqOrderService(StoreOrderService):

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: OrderIn) -> dict:
        project_data, lims_map = self.lims.process_lims(lims_order=order, new_samples=order.samples)
        status_data: dict = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["samples"], lims_map=lims_map)
        new_samples: list[Sample] = self.store_items_in_status(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"] if project_data else datetime.now(),
            ticket_id=order.ticket,
            items=status_data["samples"],
        )
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input for microbial samples."""
        return {
            "customer": order.customer,
            "order": order.name,
            "comment": order.comment,
            "samples": [
                {
                    "application": sample.application,
                    "comment": sample.comment,
                    "internal_id": sample.internal_id,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "volume": sample.volume,
                    "control": sample.control,
                }
                for sample in order.samples
            ],
        }

    def store_items_in_status(
        self,
        customer_id: str,
        order: str,
        ordered: datetime,
        ticket_id: str,
        items: list[dict],
    ) -> list[Sample]:
        customer: Customer = self._get_customer(customer_id)
        new_samples: list[Sample] = []
        status_db_order = Order(
            customer=customer,
            order_date=datetime.now(),
            ticket_id=int(ticket_id),
        )
        for sample in items:
            case_name: str = f'{sample["name"]}-case'
            case: Case = self._create_case_for_sample(
                sample=sample, customer=customer, case_name=case_name, ticket_id=ticket_id
            )
            db_sample: Sample = self._create_sample(
                sample_dict=sample,
                order=order,
                ordered=ordered,
                ticket_id=ticket_id,
                customer=customer,
            )
            db_sample = self._add_application_to_sample(
                sample=db_sample, application_tag=sample["application"]
            )
            case_sample: CaseSample = self.status_db.relate_sample(
                case=case, sample=db_sample, status=StatusEnum.unknown
            )
            status_db_order.cases.append(case)
            self.status_db.add_multiple_items_to_store([case, db_sample, case_sample])
            new_samples.append(db_sample)
        self.status_db.session.add(status_db_order)
        self.status_db.commit_to_store()
        return new_samples

    def _get_customer(self, customer_id: str) -> Customer:
        if customer := self.status_db.get_customer_by_internal_id(customer_id):
            return customer
        raise EntryNotFoundError(f"could not find customer: {customer_id}")

    def _create_case_for_sample(
        self, sample: dict, customer: Customer, case_name: str, ticket_id: str
    ) -> Case:
        if self.status_db.get_case_by_name_and_customer(case_name=case_name, customer=customer):
            raise ValueError(f"Case already exists: {case_name}.")
        case: Case = self.status_db.add_case(
            data_analysis=Workflow.RAW_DATA,
            data_delivery=DataDelivery.FASTQ,
            name=case_name,
            priority=sample["priority"],
            ticket=ticket_id,
        )
        case.customer = customer
        return case

    def _create_sample(
        self, sample_dict: dict, order, ordered, ticket_id: str, customer: Customer
    ) -> Sample:

        return self.status_db.add_sample(
            name=sample_dict["name"],
            customer=customer,
            sex=SexOptions.UNKNOWN,
            comment=sample_dict["comment"],
            internal_id=sample_dict["internal_id"],
            order=order,
            ordered=ordered,
            original_ticket=ticket_id,
            priority=sample_dict["priority"],
        )

    def _add_application_to_sample(self, sample: Sample, application_tag: str) -> Sample:
        if application_version := self.status_db.get_current_application_version_by_tag(
            tag=application_tag
        ):
            sample.application_version = application_version
            return sample
        raise EntryNotFoundError(f"Invalid application: {application_tag}")



from datetime import datetime

from cg.constants import DataDelivery, Workflow
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.services.orders.store_order_services.store_microbial_fastq_order_service import (
    StoreMicrobialFastqOrderService,
)
from cg.store.models import Case, Sample


def test_microbial_samples_to_status(
    microbial_fastq_order_to_submit: dict,
    store_microbial_fastq_order_service: StoreMicrobialFastqOrderService,
):
    # GIVEN microbial order with three samples
    order = OrderIn.parse_obj(microbial_fastq_order_to_submit, OrderType.MICROBIAL_FASTQ)

    # WHEN parsing for status
    data = store_microbial_fastq_order_service.order_to_status(order=order)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 2
    assert data["customer"] == "cust002"
    assert data["order"] == "Microbial Fastq order"
    assert data["comment"] == ""

    # THEN first sample should contain all the relevant data from the microbial order
    sample_data = data["samples"][0]
    assert sample_data["priority"] == "priority"
    assert sample_data["name"] == "prov1"
    assert sample_data.get("internal_id") is None
    assert sample_data["application"] == "WGSPCFC060"
    assert sample_data["comment"] == "sample comment"
    assert sample_data["volume"] == "1"
    assert sample_data["data_analysis"] == Workflow.MICROSALT
    assert sample_data["data_delivery"] == str(DataDelivery.FASTQ)


def test_store_samples(
    microbial_fastq_status_data: dict,
    ticket_id: str,
    store_microbial_fastq_order_service: StoreMicrobialFastqOrderService,
):
    # GIVEN a basic store with no samples and a fastq order

    assert not store_microbial_fastq_order_service.status_db._get_query(table=Sample).first()
    assert store_microbial_fastq_order_service.status_db._get_query(table=Case).count() == 0

    # WHEN storing the order
    new_samples = store_microbial_fastq_order_service.store_items_in_status(
        customer_id=microbial_fastq_status_data["customer"],
        order=microbial_fastq_status_data["order"],
        ordered=datetime.now(),
        ticket_id=ticket_id,
        items=microbial_fastq_status_data["samples"],
    )

    # THEN it should store the samples and create a case for each sample
    assert len(new_samples) == 2
    assert len(store_microbial_fastq_order_service.status_db._get_query(table=Sample).all()) == 2
    assert store_microbial_fastq_order_service.status_db._get_query(table=Case).count() == 2
    first_sample = new_samples[0]
    assert len(first_sample.links) == 1
    case_link = first_sample.links[0]
    assert case_link.case in store_microbial_fastq_order_service.status_db.get_cases()
    assert case_link.case.data_analysis
    assert case_link.case.data_delivery == DataDelivery.FASTQ



import pytest

from cg.constants import Workflow
from cg.services.orders.order_service.models import OrderQueryParams


@pytest.mark.parametrize(
    "workflow, expected_workflows",
    [
        ([Workflow.BALSAMIC], [Workflow.BALSAMIC, Workflow.BALSAMIC_UMI, Workflow.BALSAMIC_QC]),
        ([Workflow.MIP_DNA], [Workflow.MIP_DNA]),
    ],
)
def test_order_query_params_field_validator(workflow: Workflow, expected_workflows: list[Workflow]):
    """Test that the field validator expands the BALSAMIC workflow to include the UMI and QC workflows"""

    # GIVEN a order query params with BALSAMIC as workflow
    order_query_params = OrderQueryParams(workflows=workflow)

    # WHEN the field validator is called

    # THEN the BALSAMIC workflow should be expanded to include UMI and QC workflows
    assert order_query_params.workflows == expected_workflows



"""Tests for the upload Genotypes API."""

from pathlib import Path

from cg.constants import Workflow
from cg.constants.constants import SexOptions
from cg.constants.observations import MipDNAObservationsAnalysisTag
from cg.meta.upload.gt import UploadGenotypesAPI
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables
from cg.store.models import Analysis


def test_get_analysis_sex(case_qc_metrics_deliverables: Path, genotype_analysis_sex: dict):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI some qcmetrics data

    # WHEN fetching the predicted sex by the analysis
    sex: dict = UploadGenotypesAPI._get_analysis_sex_mip_dna(
        qc_metrics_file=case_qc_metrics_deliverables
    )

    # THEN assert that the the predicted sex per sample_id is returned
    assert sex == genotype_analysis_sex


def test_get_analysis_sex_raredisease(
    case_qc_metrics_deliverables_raredisease: Path, sample_id: str
):
    """Test to get the predicted sex from a MIP run using the upload Genotypes API"""
    # GIVEN a UploadGenotypesAPI some qcmetrics data

    # WHEN fetching the predicted sex by the analysis
    sex: str = UploadGenotypesAPI._get_analysis_sex_raredisease(
        UploadGenotypesAPI,
        qc_metrics_file=case_qc_metrics_deliverables_raredisease,
        sample_id=sample_id,
    )

    # THEN assert that the the predicted sex per sample_id is returned
    assert sex == SexOptions.MALE


def test_get_parsed_qc_metrics_deliverables_mip(case_qc_metrics_deliverables: Path):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI and the path to a qc_metrics file with case data

    # WHEN fetching the predicted sex
    metrics_object: MIPMetricsDeliverables = (
        UploadGenotypesAPI._get_parsed_qc_metrics_deliverables_mip_dna(case_qc_metrics_deliverables)
    )
    # THEN assert that it was successfully created
    assert isinstance(metrics_object, MIPMetricsDeliverables)


def test_get_parsed_qc_metrics_deliverables_raredisease(case_qc_metrics_deliverables: Path):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI and the path to a QC metrics file with case data

    # WHEN fetching the predicted sex
    metrics_object = UploadGenotypesAPI._get_parsed_qc_metrics_deliverables_raredisease(
        case_qc_metrics_deliverables
    )

    # First, check if the object is a list
    assert isinstance(metrics_object, list)

    # Then, check if all items in the list are instances of MetricsBase
    assert all(isinstance(item, MetricsBase) for item in metrics_object)


def test_get_genotype_file(upload_genotypes_api: UploadGenotypesAPI, case_id: str):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI populated with some data in housekeeper

    # WHEN fetching the gbcf file with the api
    gbcf = upload_genotypes_api._get_genotype_file(case_id)

    # THEN assert that the file has the correct tag
    assert MipDNAObservationsAnalysisTag.PROFILE_GBCF in (tag.name for tag in gbcf.tags)


def test_get_data_mip(
    analysis_obj: Analysis,
    genotype_analysis_sex: dict,
    mocker,
    upload_genotypes_api: UploadGenotypesAPI,
):
    """Test to get data from the UploadGenotypesAPI"""
    # GIVEN a UploadGenotypeAPI populated with some data
    # GIVEN an analysis object with a trio

    # GIVEN analysis sex were generated and could be found
    mocker.patch.object(UploadGenotypesAPI, "_get_analysis_sex_mip_dna")
    UploadGenotypesAPI._get_analysis_sex_mip_dna.return_value = genotype_analysis_sex

    analysis_obj.workflow = Workflow.MIP_DNA

    # WHEN parsing the data
    result = upload_genotypes_api.get_genotype_data(analysis=analysis_obj)

    # THEN assert that the the number of samples sex is set
    assert len(result["samples_sex"]) == 3


def test_get_data_raredisease(
    analysis_obj: Analysis,
    genotype_analysis_sex: dict,
    mocker,
    upload_genotypes_api: UploadGenotypesAPI,
):
    """Test to get data from the UploadGenotypesAPI"""
    # GIVEN a UploadGenotypeAPI populated with some data
    # GIVEN an analysis object with a trio

    # GIVEN analysis sex were generated and could be found
    mocker.patch.object(UploadGenotypesAPI, "_get_analysis_sex_raredisease")
    UploadGenotypesAPI._get_analysis_sex_raredisease.return_value = genotype_analysis_sex

    analysis_obj.workflow = Workflow.RAREDISEASE

    # WHEN parsing the data
    result = upload_genotypes_api.get_genotype_data(analysis=analysis_obj)

    # THEN assert that the the number of samples sex is set
    assert len(result["samples_sex"]) == 3



from pathlib import Path

import pytest


@pytest.fixture
def csv_file_path(fixtures_dir: Path) -> Path:
    """Return a file path to example CSV file."""
    return Path(fixtures_dir, "io", "example.csv")


@pytest.fixture
def csv_stream() -> str:
    """Return string with CSV format."""
    return """Lorem,ipsum,sit,amet"""


@pytest.fixture
def csv_temp_path(cg_dir: Path) -> Path:
    """Return a temp file path to use when writing csv."""
    return Path(cg_dir, "write.csv")



import logging
from pathlib import Path

from cg.constants.delivery import INBOX_NAME
from cg.services.deliver_files.file_fetcher.models import (
    DeliveryFiles,
    DeliveryMetaData,
)
from cg.services.deliver_files.file_mover.abstract import DestinationFilesMover
from cg.services.deliver_files.utils import FileMover

LOG = logging.getLogger(__name__)


class CustomerInboxDestinationFilesMover(DestinationFilesMover):
    """
    Class to move files to the customer folder.
    """

    def __init__(self, file_mover: FileMover):
        self.file_mover = file_mover

    def move_files(self, delivery_files: DeliveryFiles, delivery_base_path: Path) -> DeliveryFiles:
        """
        Move the files to the customer folder.
        args:
            delivery_files: DeliveryFiles: The files to move.
            delivery_base_path: Path: The path to move the files to.
        """
        inbox_ticket_dir_path: Path = self._create_ticket_inbox_dir_path(
            delivery_base_path=delivery_base_path, delivery_data=delivery_files.delivery_data
        )
        delivery_files.delivery_data.delivery_path = inbox_ticket_dir_path

        self.file_mover.create_directories(
            base_path=delivery_base_path,
            directories={str(inbox_ticket_dir_path.relative_to(delivery_base_path))},
        )
        delivery_files.case_files = self.file_mover.move_and_update_files(
            file_models=delivery_files.case_files, target_dir=inbox_ticket_dir_path
        )
        delivery_files.sample_files = self.file_mover.move_and_update_files(
            file_models=delivery_files.sample_files, target_dir=inbox_ticket_dir_path
        )
        return delivery_files

    @staticmethod
    def _create_ticket_inbox_dir_path(
        delivery_base_path: Path, delivery_data: DeliveryMetaData
    ) -> Path:
        """Generate the path to the ticket inbox directory.
        args:
            delivery_base_path: The base path to the delivery folder.
            delivery_data: The delivery data containing the customer internal id and ticket id.
        """
        return Path(
            delivery_base_path,
            delivery_data.customer_internal_id,
            INBOX_NAME,
            delivery_data.ticket_id,
        )



"""Module for Tomte Analysis API."""

import logging
from pathlib import Path

from cg.constants import Workflow
from cg.constants.constants import Strandedness
from cg.constants.nf_analysis import TOMTE_METRIC_CONDITIONS
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.tomte.tomte import TomteParameters, TomteSampleSheetEntry, TomteSampleSheetHeaders
from cg.resources import TOMTE_BUNDLE_FILENAMES_PATH
from cg.store.models import CaseSample

LOG = logging.getLogger(__name__)


class TomteAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Tomte processes and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.TOMTE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.tomte.root
        self.workflow_bin_path: str = config.tomte.workflow_bin_path
        self.profile: str = config.tomte.profile
        self.conda_env: str = config.tomte.conda_env
        self.conda_binary: str = config.tomte.conda_binary
        self.platform: str = config.tomte.platform
        self.params: str = config.tomte.params
        self.workflow_config_path: str = config.tomte.config
        self.resources: str = config.tomte.resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.tomte.tower_workflow
        self.account: str = config.tomte.slurm.account
        self.email: str = config.tomte.slurm.mail_user
        self.compute_env_base: str = config.tomte.compute_env
        self.revision: str = config.tomte.revision
        self.nextflow_binary_path: str = config.tomte.binary_path

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return TomteSampleSheetHeaders.list()

    @property
    def is_gene_panel_required(self) -> bool:
        """Return True if a gene panel is needs to be created using the information in StatusDB and exporting it from Scout."""
        return True

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return path to bundle template."""
        return TOMTE_BUNDLE_FILENAMES_PATH

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        fastq_forward_read_paths, fastq_reverse_read_paths = self.get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = TomteSampleSheetEntry(
            case_id=case_sample.case.internal_id,
            name=case_sample.sample.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            strandedness=Strandedness.REVERSE,
        )
        return sample_sheet_entry.reformat_sample_content

    def get_built_workflow_parameters(self, case_id: str) -> TomteParameters:
        """Return parameters."""
        return TomteParameters(
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            gene_panel_clinical_filter=self.get_gene_panels_path(case_id=case_id),
            tissue=self.get_case_source_type(case_id=case_id),
            genome=self.get_genome_build(case_id=case_id),
        )

    def get_workflow_metrics(self, metric_id: str) -> dict:
        return TOMTE_METRIC_CONDITIONS



class FreshdeskException(Exception):
    """Base exception for Freshdesk errors."""

    pass


class FreshdeskAPIException(FreshdeskException):
    """The Freshdesk API returned an error."""

    pass


class FreshdeskModelException(FreshdeskException):
    """Pydantic model validation failed."""

    pass



"""Helper functions related to the genome build."""

from cg.constants.constants import GenomeVersion
from cg.constants.scout import GenomeBuild
from cg.exc import CgError
from cg.store.models import Case


def get_genome_build(case: Case) -> GenomeVersion:
    """Return reference genome version for a case.
    Raises CgError if this information is missing or inconsistent for the samples linked to a case.
    """
    reference_genome: set[str] = {sample.reference_genome for sample in case.samples}
    if len(reference_genome) == 1:
        return reference_genome.pop()
    if len(reference_genome) > 1:
        raise CgError(
            f"Samples linked to case {case.internal_id} have different reference genome versions set"
        )
    raise CgError(f"No reference genome specified for case {case.internal_id}")


def genome_to_scout_format(genome: GenomeVersion) -> str:
    mapping = {
        GenomeVersion.GRCh37: GenomeBuild.hg19,
        GenomeVersion.GRCh38: GenomeBuild.hg38,
        GenomeVersion.HG19: GenomeBuild.hg19,
        GenomeVersion.HG38: GenomeBuild.hg38,
    }
    # Check if the current instance is in the mapping, raise an error if not
    if genome not in mapping:
        raise ValueError(f"Genome build incompatible with scout: {genome.value}")
    return mapping[genome]



from cg.models.fohm.reports import FohmComplementaryReport, FohmPangolinReport


def test_instantiate_fohm_complementary_report(fohm_complementary_report_raw: dict[str, str]):
    """Tests report dict against a pydantic FohmComplementaryReport."""

    # GIVEN report dicts

    # WHEN instantiating a FOHM report object
    fohm_report = FohmComplementaryReport.model_validate(fohm_complementary_report_raw)

    # THEN assert that it was successfully created
    assert isinstance(fohm_report, FohmComplementaryReport)


def test_instantiate_fohm_pangolin_report(fohm_pangolin_report_raw: dict[str, str]):
    """Tests report dict against a pydantic FohmPangolinReport."""

    # GIVEN report dicts

    # WHEN instantiating a FOHM report object
    fohm_report = FohmPangolinReport.model_validate(fohm_pangolin_report_raw)

    # THEN assert that it was successfully created
    assert isinstance(fohm_report, FohmPangolinReport)



from datetime import datetime
from unittest.mock import Mock

from cg.constants import Workflow
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.store.models import Analysis
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_get_analyses_to_upload_for_workflow(
    helpers: StoreHelpers, base_store: Store, timestamp_now: datetime
):
    # GIVEN an analysis service and a store with analyses to upload and not to upload
    analysis_upload: Analysis = helpers.add_analysis(
        store=base_store,
        uploaded_at=None,
        upload_started=None,
        workflow=Workflow.RAW_DATA,
        completed_at=timestamp_now,
    )
    new_case = helpers.add_case(store=base_store, name="no_upload_case")
    analysis_no_upload: Analysis = helpers.add_analysis(
        store=base_store,
        case=new_case,
        uploaded_at=timestamp_now,
        upload_started=None,
        workflow=Workflow.RAW_DATA,
        completed_at=timestamp_now,
    )
    analysis_service = AnalysisService(analysis_client=Mock(), status_db=base_store)

    # WHEN getting analyses to upload
    analyses: list[Analysis] = analysis_service.get_analyses_to_upload_for_workflow(
        workflow=Workflow.RAW_DATA
    )

    # THEN only the analyses to upload should be returned
    assert analyses == [analysis_upload]
    assert analysis_no_upload not in analyses



"""Module for demultiplex fixtures returning strings."""

import pytest


@pytest.fixture
def flow_cell_name() -> str:
    """Return flow cell name."""
    return "HVKJCDRXX"


@pytest.fixture
def flow_cell_full_name(flow_cell_name: str) -> str:
    """Return flow cell full name."""
    return "201203_D00483_0200_AHVKJCDRXX"


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_id() -> str:
    """Return the id of a HiSeqX flow cell with only one index."""
    return "HJCFFALXX"


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_name(hiseq_x_single_index_flow_cell_id) -> str:
    """Return the full name of a HiSeqX flow cell with only one index."""
    return f"170517_ST-E00266_0210_B{hiseq_x_single_index_flow_cell_id}"


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_id() -> str:
    """Return the id of a HiSeqX flow cell with double indexes."""
    return "HL32LCCXY"


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_name(hiseq_x_dual_index_flow_cell_id: str) -> str:
    """Return the full name of a HiSeqX flow cell with double indexes."""
    return f"180508_ST-E00269_0269_A{hiseq_x_dual_index_flow_cell_id}"


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_id() -> str:
    """Return the id of a HiSeq2500 flow cell with double indexes."""
    return "HM2LNBCX2"


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_name(hiseq_2500_dual_index_flow_cell_id: str) -> str:
    """Return the full name of a HiSeq2500 flow cell with double indexes."""
    return f"181005_D00410_0735_B{hiseq_2500_dual_index_flow_cell_id}"


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_id() -> str:
    """Return the id of a HiSeq2500 flow cell with custom indexes."""
    return "HGYFNBCX2"


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_name(hiseq_2500_custom_index_flow_cell_id) -> str:
    """Return the full name of a HiSeq2500 flow cell with custom indexes."""
    return f"180509_D00450_0598_B{hiseq_2500_custom_index_flow_cell_id}"


@pytest.fixture(scope="session")
def novaseq_6000_pre_1_5_kits_flow_cell_id() -> str:
    """Return the id of a pre-1.5 kits NovaSeq6000 flow cell."""
    return "HLYWYDSXX"


@pytest.fixture(scope="session")
def novaseq_6000_pre_1_5_kits_flow_cell_full_name(
    novaseq_6000_pre_1_5_kits_flow_cell_id: str,
) -> str:
    """Return the full name of a pre-1.5 kits NovaSeq6000 flow cell."""
    return f"190927_A00689_0069_B{novaseq_6000_pre_1_5_kits_flow_cell_id}"


@pytest.fixture(scope="session")
def novaseq_6000_post_1_5_kits_flow_cell_id() -> str:
    """Return the id of a post-1.5 kits NovaSeq6000 flow cell."""
    return "HK33MDRX3"


@pytest.fixture(scope="session")
def novaseq_6000_post_1_5_kits_flow_cell_full_name(
    novaseq_6000_post_1_5_kits_flow_cell_id: str,
) -> str:
    """Return the full name of a post-1.5 kits NovaSeq6000 flow cell."""
    return f"230912_A00187_1009_A{novaseq_6000_post_1_5_kits_flow_cell_id}"


@pytest.fixture(scope="session")
def novaseq_x_flow_cell_id() -> str:
    """Return the id of a NovaSeqX flow cell."""
    return "22F52TLT3"


@pytest.fixture(scope="session")
def novaseq_x_flow_cell_full_name(novaseq_x_flow_cell_id: str) -> str:
    """Return the full name of a NovaSeqX flow cell."""
    return f"20231108_LH00188_0028_B{novaseq_x_flow_cell_id}"


@pytest.fixture(scope="session")
def seven_canonical_flow_cell_ids(
    hiseq_x_single_index_flow_cell_id: str,
    hiseq_x_dual_index_flow_cell_id: str,
    hiseq_2500_dual_index_flow_cell_id: str,
    hiseq_2500_custom_index_flow_cell_id: str,
    novaseq_6000_pre_1_5_kits_flow_cell_id: str,
    novaseq_6000_post_1_5_kits_flow_cell_id: str,
    novaseq_x_flow_cell_id: str,
) -> list[str]:
    """Return a list of seven canonical flow cell ids."""
    return [
        hiseq_x_single_index_flow_cell_id,
        hiseq_x_dual_index_flow_cell_id,
        hiseq_2500_dual_index_flow_cell_id,
        hiseq_2500_custom_index_flow_cell_id,
        novaseq_6000_pre_1_5_kits_flow_cell_id,
        novaseq_6000_post_1_5_kits_flow_cell_id,
        novaseq_x_flow_cell_id,
    ]


@pytest.fixture(scope="session")
def demultiplex_log_file_names(novaseq_6000_post_1_5_kits_flow_cell_id: str) -> list[str]:
    """Return a list of demultiplex log file names."""
    return [
        f"{novaseq_6000_post_1_5_kits_flow_cell_id}_demultiplex.stderr",
        f"{novaseq_6000_post_1_5_kits_flow_cell_id}_demultiplex.stdout",
    ]


@pytest.fixture
def novaseq_6000_post_1_5_kits_fastq_file_names(
    selected_novaseq_6000_post_1_5_kits_sample_ids: list[str],
) -> list[str]:
    """Return a list of fastq file names for a NovaSeq6000 post 1.5 kits flow cell."""
    fastq_files: list[str] = []
    lanes = [1, 2]
    for sample_id, lane in zip(selected_novaseq_6000_post_1_5_kits_sample_ids, lanes):
        fastq_files.extend(
            [
                f"{sample_id}_S1_L00{lane}_R1_001.fastq.gz",
                f"{sample_id}_S1_L00{lane}_R2_001.fastq.gz",
            ]
        )
    return fastq_files


@pytest.fixture
def novaseq_6000_post_1_5_kits_fastq_file_lane_1(
    novaseq_6000_post_1_5_kits_fastq_file_names: list[str],
) -> str:
    """Return the fastq file name for lane 1."""
    return novaseq_6000_post_1_5_kits_fastq_file_names[0]


@pytest.fixture
def novaseq_6000_post_1_5_kits_fastq_file_lane_1_with_flow_cell_id(
    novaseq_6000_post_1_5_kits_fastq_file_lane_1: str,
    novaseq_6000_post_1_5_kits_flow_cell_id: str,
) -> str:
    """Return the fastq file name for lane 1 with flow cell id."""
    return (
        f"{novaseq_6000_post_1_5_kits_flow_cell_id}_{novaseq_6000_post_1_5_kits_fastq_file_lane_1}"
    )



class ArnoldAPIError(Exception):
    def __init__(self, message: str = ""):
        super(ArnoldAPIError, self).__init__()
        self.message = message


class ArnoldClientError(ArnoldAPIError):
    """Raises an error in case of an error when parsing a JSON file."""


class ArnoldServerError(ArnoldAPIError):
    """Raises an error in case of an error when parsing a JSON file."""



from sqlalchemy.orm import Query

from cg.constants import Workflow
from cg.store.filters.status_application_limitations_filters import (
    filter_application_limitations_by_tag,
    filter_application_limitations_by_workflow,
)
from cg.store.store import Store
from tests.store.conftest import StoreConstants


def test_filter_application_limitations_by_tag(
    store_with_application_limitations: Store,
    tag=StoreConstants.TAG_APPLICATION_WITH_ATTRIBUTES.value,
) -> None:
    """Test to get application limitations by tag."""

    # GIVEN a store with application limitations

    # WHEN getting an application limitations by tag
    application_limitations: Query = filter_application_limitations_by_tag(
        application_limitations=store_with_application_limitations._get_join_application_limitations_query(),
        tag=tag,
    )

    # ASSERT that application limitations is a query
    assert isinstance(application_limitations, Query)

    # THEN assert that the application limitations was found
    assert (
        application_limitations.all()
        and len(application_limitations.all()) == 1
        and application_limitations.all()[0].application.tag == tag
    )


def test_filter_application_limitations_by_workflow(
    store_with_application_limitations: Store,
    workflow=Workflow.BALSAMIC,
) -> None:
    """Test to get application limitations by workflow."""

    # GIVEN a store with application limitations

    # WHEN getting an application limitations by workflow
    application_limitations: Query = filter_application_limitations_by_workflow(
        application_limitations=store_with_application_limitations._get_join_application_limitations_query(),
        workflow=workflow,
    )

    # ASSERT that application limitations is a query
    assert isinstance(application_limitations, Query)

    # THEN assert the application limitations was found
    assert (
        application_limitations.all()
        and len(application_limitations.all()) == 1
        and application_limitations.all()[0].workflow == workflow
    )



import logging
from datetime import datetime

from cg.constants import DataDelivery, Workflow
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import SexEnum, StatusEnum
from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.services.orders.submitters.order_submitter import StoreOrderService
from cg.store.models import ApplicationVersion, CaseSample, Customer, Order, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StorePacBioOrderService(StoreOrderService):
    """Storing service for PacBio Long Read orders."""

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for PacBio Long Read delivery."""

        project_data, lims_map = self.lims.process_lims(lims_order=order, new_samples=order.samples)
        status_data: dict = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["samples"], lims_map=lims_map)
        new_samples = self._store_samples_in_statusdb(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket_id=order.ticket,
            samples=status_data["samples"],
        )
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for PacBio-only orders."""
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "samples": [
                {
                    "application": sample.application,
                    "comment": sample.comment,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "sex": sample.sex,
                    "tumour": sample.tumour,
                    "volume": sample.volume,
                    "subject_id": sample.subject_id,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def _store_samples_in_statusdb(
        self, customer_id: str, order: str, ordered: datetime, ticket_id: str, samples: list[dict]
    ) -> list[Sample]:
        """Store PacBio samples and cases in the StatusDB."""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        status_db_order = Order(
            customer=customer,
            order_date=datetime.now(),
            ticket_id=int(ticket_id),
        )
        new_samples = []
        with self.status_db.session.no_autoflush:
            for sample in samples:
                sample_name: str = sample["name"]
                new_sample = self.status_db.add_sample(
                    name=sample_name,
                    sex=sample["sex"] or SexEnum.unknown,
                    comment=sample["comment"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket_id,
                    priority=sample["priority"],
                    tumour=sample["tumour"],
                    subject_id=sample["subject_id"],
                )
                new_sample.customer = customer
                application_tag: str = sample["application"]
                application_version: ApplicationVersion = (
                    self.status_db.get_current_application_version_by_tag(tag=application_tag)
                )
                new_sample.application_version = application_version
                new_samples.append(new_sample)
                case = self.status_db.add_case(
                    data_analysis=Workflow(sample["data_analysis"]),
                    data_delivery=DataDelivery(sample["data_delivery"]),
                    name=f"{sample_name}-case",
                    priority=sample["priority"],
                    ticket=ticket_id,
                )
                case.customer = customer
                new_relationship: CaseSample = self.status_db.relate_sample(
                    case=case, sample=new_sample, status=StatusEnum.unknown
                )
                status_db_order.cases.append(case)
                self.status_db.session.add_all([case, new_relationship])

        self.status_db.session.add(status_db_order)
        self.status_db.session.add_all(new_samples)
        self.status_db.session.commit()
        return new_samples



from datetime import datetime

import pytest

from cg.constants import SequencingRunDataAvailability
from cg.constants.constants import ControlOptions
from cg.constants.sequencing import Sequencers
from cg.store.models import Analysis, IlluminaSampleSequencingMetrics, IlluminaSequencingRun, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_update_illumina_sequencing_run_availability(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with Illumina Sequencing Runs that have data availability ON_DISK
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_runs_by_data_availability(
            [SequencingRunDataAvailability.ON_DISK]
        )[0]
    )
    assert sequencing_run.data_availability == SequencingRunDataAvailability.ON_DISK

    # WHEN updating the data availability of a sequencing run
    store_with_illumina_sequencing_data.update_illumina_sequencing_run_data_availability(
        sequencing_run=sequencing_run, data_availability=SequencingRunDataAvailability.REQUESTED
    )

    # THEN the data availability of the sequencing run is updated
    assert sequencing_run.data_availability == SequencingRunDataAvailability.REQUESTED


def test_update_illumina_sequencing_run_has_backup(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with Illumina Sequencing Runs that have backup status False
    sequencing_run: IlluminaSequencingRun = store_with_illumina_sequencing_data._get_query(
        IlluminaSequencingRun
    ).all()[0]
    assert sequencing_run.has_backup is False

    # WHEN updating the backup status of a sequencing run
    store_with_illumina_sequencing_data.update_illumina_sequencing_run_has_backup(
        sequencing_run=sequencing_run, has_backup=True
    )

    # THEN the backup status of the sequencing run is updated
    assert sequencing_run.has_backup is True


def test_update_sample_reads_illumina(
    store_with_illumina_sequencing_data: Store, selected_novaseq_x_sample_ids: list[str]
):
    # GIVEN a store with Illumina Sequencing Runs and a sample id
    sample: Sample = store_with_illumina_sequencing_data.get_sample_by_internal_id(
        selected_novaseq_x_sample_ids[0]
    )
    assert sample.reads == 0

    # WHEN updating the sample reads for a sequencing run
    store_with_illumina_sequencing_data.update_sample_reads_illumina(
        internal_id=selected_novaseq_x_sample_ids[0], sequencer_type=Sequencers.NOVASEQX
    )

    # THEN the total reads for the sample is updated
    total_reads_for_sample: int = 0
    sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics
    for sample_metric in sample_metrics:
        total_reads_for_sample += sample_metric.total_reads_in_lane
    assert sample.reads == total_reads_for_sample


def test_update_sample_reads_illumina_fail_q30(
    store_with_illumina_sequencing_data: Store, selected_novaseq_x_sample_ids: list[str]
):
    # GIVEN a store with Illumina Sequencing Runs and a sample id
    sample: Sample = store_with_illumina_sequencing_data.get_sample_by_internal_id(
        selected_novaseq_x_sample_ids[0]
    )
    assert sample.reads == 0
    # GIVEN that the q30 for the lane is below the threshold for the sequencer type
    sample.sample_run_metrics[0].base_passing_q30_percent = 30

    # WHEN updating the sample reads for a sequencing run
    store_with_illumina_sequencing_data.update_sample_reads_illumina(
        internal_id=selected_novaseq_x_sample_ids[0], sequencer_type=Sequencers.NOVASEQX
    )

    # THEN the total reads for the sample is updated
    assert sample.reads == 0


def test_update_sample_reads_illumina_negative_control(
    store_with_illumina_sequencing_data: Store, selected_novaseq_x_sample_ids: list[str]
):
    # GIVEN a store with Illumina Sequencing Runs and a sample id
    sample: Sample = store_with_illumina_sequencing_data.get_sample_by_internal_id(
        selected_novaseq_x_sample_ids[0]
    )
    assert sample.reads == 0
    # GIVEN that the sample is a negative control sample
    sample.control = ControlOptions.NEGATIVE

    # WHEN updating the sample reads for a sequencing run
    store_with_illumina_sequencing_data.update_sample_reads_illumina(
        internal_id=selected_novaseq_x_sample_ids[0], sequencer_type=Sequencers.NOVASEQX
    )

    # THEN the total reads for the sample is updated
    total_reads_for_sample: int = 0
    sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics
    for sample_metric in sample_metrics:
        total_reads_for_sample += sample_metric.total_reads_in_lane
    assert sample.reads == total_reads_for_sample


def test_update_sample_reads_pacbio(
    pacbio_barcoded_sample_internal_id: str,
    store: Store,
    helpers: StoreHelpers,
):
    """Test updating the reads for a PacBio sample."""
    # GIVEN a store with a PacBio sample
    sample: Sample = helpers.add_sample(store=store, internal_id=pacbio_barcoded_sample_internal_id)
    assert sample
    assert sample.reads == 0
    reads: int = 10000

    # WHEN updating the reads for the sample
    store.update_sample_reads(internal_id=pacbio_barcoded_sample_internal_id, reads=reads)

    # THEN the reads for the sample is updated
    sample: Sample = store.get_sample_by_internal_id(pacbio_barcoded_sample_internal_id)
    assert sample.reads == reads


@pytest.mark.parametrize(
    "sample_id_fixture",
    ["sample_id_sequenced_on_multiple_flow_cells", "pacbio_barcoded_sample_internal_id"],
    ids=["Illumina", "Pacbio"],
)
def test_update_sample_last_sequenced_at(
    store: Store,
    helpers: StoreHelpers,
    sample_id_fixture: str,
    timestamp_now: datetime,
    request: pytest.FixtureRequest,
):
    """Test updating the last sequenced at date for a sample."""
    # GIVEN a store with a sample
    sample_id: str = request.getfixturevalue(sample_id_fixture)
    sample: Sample = helpers.add_sample(store=store, internal_id=sample_id)
    assert sample.last_sequenced_at is None

    # WHEN updating the last sequenced at date for a sample
    store.update_sample_sequenced_at(internal_id=sample_id, date=timestamp_now)

    # THEN the last sequenced at date for the sample is updated
    sample: Sample = store.get_sample_by_internal_id(sample_id)
    assert sample.last_sequenced_at == timestamp_now


def test_update_analysis_uploaded_at(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store, timestamp_yesterday: datetime
):
    # GIVEN a store with an analysis
    analysis: Analysis = store_with_analyses_for_cases_not_uploaded_fluffy._get_query(
        Analysis
    ).first()
    assert analysis.uploaded_at != timestamp_yesterday

    # WHEN updating the uploaded_at field
    store_with_analyses_for_cases_not_uploaded_fluffy.update_analysis_uploaded_at(
        analysis_id=analysis.id, uploaded_at=timestamp_yesterday
    )

    # THEN the uploaded_at field is updated
    assert analysis.uploaded_at == timestamp_yesterday


def test_update_analysis_upload_started_at(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store, timestamp_yesterday: datetime
):
    # GIVEN a store with an analysis
    analysis: Analysis = store_with_analyses_for_cases_not_uploaded_fluffy._get_query(
        Analysis
    ).first()
    assert analysis.upload_started_at != timestamp_yesterday

    # WHEN updating the upload_started_at field
    store_with_analyses_for_cases_not_uploaded_fluffy.update_analysis_upload_started_at(
        analysis_id=analysis.id, upload_started_at=timestamp_yesterday
    )

    # THEN the upload_started_at field is updated
    assert analysis.upload_started_at == timestamp_yesterday



import datetime as dt

from cg.constants import DataDelivery, Workflow
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.store_order_services.store_pool_order import StorePoolOrderService
from cg.store.models import Case, Pool, Sample
from cg.store.store import Store


def test_pools_to_status(
    rml_order_to_submit: dict, store_pool_order_service: StorePoolOrderService
):
    # GIVEN a rml order with three samples in one pool
    order = OrderIn.parse_obj(rml_order_to_submit, OrderType.RML)

    # WHEN parsing for status
    data = store_pool_order_service.order_to_status(order=order)

    # THEN it should pick out the general information
    assert data["customer"] == "cust000"
    assert data["order"] == "#123456"
    assert data["comment"] == "order comment"

    # ... and information about the pool(s)
    assert len(data["pools"]) == 2
    pool = data["pools"][0]
    assert pool["name"] == "pool-1"
    assert pool["application"] == "RMLP05R800"
    assert pool["data_analysis"] == Workflow.RAW_DATA
    assert pool["data_delivery"] == str(DataDelivery.FASTQ)
    assert len(pool["samples"]) == 2
    sample = pool["samples"][0]
    assert sample["name"] == "sample1"
    assert sample["comment"] == "test comment"
    assert pool["priority"] == "research"
    assert sample["control"] == "negative"


def test_store_rml(
    base_store: Store,
    rml_status_data: dict,
    ticket_id: str,
    store_pool_order_service: StorePoolOrderService,
):
    # GIVEN a basic store with no samples and a rml order
    assert base_store._get_query(table=Pool).count() == 0
    assert base_store._get_query(table=Case).count() == 0
    assert not base_store._get_query(table=Sample).first()

    # WHEN storing the order
    new_pools = store_pool_order_service.store_items_in_status(
        customer_id=rml_status_data["customer"],
        order=rml_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=rml_status_data["pools"],
    )

    # THEN it should update the database with new pools
    assert len(new_pools) == 2

    assert base_store._get_query(table=Pool).count() == base_store._get_query(table=Case).count()
    assert len(base_store._get_query(table=Sample).all()) == 4

    # ASSERT that there is one negative sample
    negative_samples = 0
    for sample in base_store._get_query(table=Sample).all():
        if sample.control == "negative":
            negative_samples += 1
    assert negative_samples == 1

    new_pool = base_store._get_query(table=Pool).order_by(Pool.created_at.desc()).first()
    assert new_pool == new_pools[1]

    assert new_pool.name == "pool-2"
    assert new_pool.application_version.application.tag == "RMLP05R800"
    assert not hasattr(new_pool, "data_analysis")

    new_case = base_store.get_cases()[0]
    assert new_case.data_analysis == Workflow.RAW_DATA
    assert new_case.data_delivery == str(DataDelivery.FASTQ)

    # and that the pool is set for invoicing but not the samples of the pool
    assert not new_pool.no_invoice
    for link in new_case.links:
        assert link.sample.no_invoice



from pathlib import Path
from pydantic import BaseModel


class FormattedFile(BaseModel):
    original_path: Path
    formatted_path: Path


class FormattedFiles(BaseModel):
    files: list[FormattedFile]



"""Tests for the dict module."""

from pathlib import Path

from cg.utils.dict import (
    get_full_path_dictionary,
    get_list_from_dictionary,
    remove_duplicate_dicts,
)


def test_remove_duplicate_dicts():
    # GIVEN a list of dicts
    dicts = [{"a": 1, "b": 2}, {"c": 1, "d": 4}, {"a": 1, "b": 2}, {"c": 1, "d": 4}]

    # WHEN removing duplicate dicts
    content: list[dict] = remove_duplicate_dicts(dicts)

    # THEN a list of dicts is returned
    assert isinstance(content[0], dict)

    # THEN duplicates are removed
    assert len(content) == 2
    assert len(content[0]) == 2
    assert len(content[1]) == 2


def test_remove_duplicate_dicts_when_no_duplicates():
    # GIVEN a list of dicts
    dicts = [{"a": 1, "b": 2}, {"c": 1, "d": 4}, {"f": 1, "b": 2}, {"c": 1, "d": 1}]

    # WHEN removing duplicate dicts
    content: list[dict] = remove_duplicate_dicts(dicts)

    # THEN a list of dicts is returned
    assert isinstance(content[0], dict)

    # THEN all dicts remain
    assert len(content) == 4
    assert len(content[3]) == 2


def test_get_list_from_dictionary():
    """Test the formatting of the flattened dictionary."""

    # GIVEN a mock dictionary
    mock_dict = {"a": 1, "b": 2, "c": None, "d": "3"}

    # GIVEN the expected output
    expected_output = ["a", 1, "b", 2, "d", "3"]

    # WHEN calling the function
    flattened_dict: dict = get_list_from_dictionary(mock_dict)

    # THEN assert the output should match the expected list
    assert flattened_dict == expected_output


def test_get_list_from_dictionary_empty_input():
    """Test the formatting of the flattened dictionary."""

    # GIVEN an empty dictionary
    mock_dict = {}

    # WHEN calling the function
    flattened_dict: dict = get_list_from_dictionary(mock_dict)

    # THEN assert the output matches the expected list
    assert not flattened_dict


def test_get_full_path_dictionary(hk_file) -> dict:
    """Test full path parsing of a dictionary."""

    # GIVEN a housekeeper file, an empty value, and a dictionary containing it
    file_dict: dict = {"file": hk_file, "empty": None}

    # WHEN obtaining the full path of the elements of a dictionary
    file_full_path_dictionary: dict = get_full_path_dictionary(file_dict)

    # THEN the obtained dictionary path should be absolute
    assert Path(file_full_path_dictionary["file"]).is_absolute()
    assert file_full_path_dictionary["empty"] is None



"""Filters for the Illumina Sample Sequencing Metrics objects."""

from enum import Enum

from sqlalchemy.orm import Query

from cg.store.models import IlluminaSampleSequencingMetrics


def filter_by_lane(metrics: Query, lane: int, **kwargs) -> Query:
    """Filter metrics by lane."""
    return metrics.filter(IlluminaSampleSequencingMetrics.flow_cell_lane == lane)


class IlluminaMetricsFilter(Enum):
    BY_LANE: callable = filter_by_lane


def apply_illumina_metrics_filter(
    metrics: Query,
    filter_functions: list[callable],
    lane: int | None = None,
) -> Query:
    for filter_function in filter_functions:
        metrics: Query = filter_function(
            metrics=metrics,
            lane=lane,
        )
    return metrics



"""Fixtures for PacBio metrics objects."""

import pytest

from cg.constants.pacbio import (
    BarcodeMetricsAliases,
    CCSAttributeAliases,
    ControlAttributeAliases,
    LoadingAttributesAliases,
    PolymeraseDataAttributeAliases,
    SampleMetricsAliases,
    SmrtLinkDatabasesAliases,
)
from cg.services.run_devices.pacbio.metrics_parser.models import (
    BarcodeMetrics,
    ControlMetrics,
    PacBioMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SampleMetrics,
    SmrtlinkDatasetsMetrics,
)


@pytest.fixture
def pac_bio_read_metrics() -> ReadMetrics:
    data: dict[str, any] = {
        CCSAttributeAliases.HIFI_READS: 7815301,
        CCSAttributeAliases.HIFI_YIELD: 115338856495,
        CCSAttributeAliases.HIFI_MEAN_READ_LENGTH: 14758,
        CCSAttributeAliases.HIFI_MEDIAN_READ_LENGTH: 14679,
        CCSAttributeAliases.HIFI_READ_LENGTH_N50: 14679,
        CCSAttributeAliases.HIFI_MEDIAN_READ_QUALITY: 35,
        CCSAttributeAliases.PERCENT_Q30: 93.10000000000001,
        CCSAttributeAliases.FAILED_READS: 469053,
        CCSAttributeAliases.FAILED_YIELD: 7404831038,
        CCSAttributeAliases.FAILED_MEAN_READ_LENGTH: 15786,
    }
    return ReadMetrics.model_validate(data)


@pytest.fixture
def pac_bio_control_metrics() -> ControlMetrics:
    data: dict[str, any] = {
        ControlAttributeAliases.NUMBER_OF_READS: 2368,
        ControlAttributeAliases.MEAN_READ_LENGTH: 69936,
        ControlAttributeAliases.PERCENT_MEAN_READ_CONCORDANCE: 0.90038,
        ControlAttributeAliases.PERCENT_MODE_READ_CONCORDANCE: 0.91,
    }
    return ControlMetrics.model_validate(data)


@pytest.fixture
def pac_bio_productivity_metrics() -> ProductivityMetrics:
    data: dict[str, any] = {
        LoadingAttributesAliases.PRODUCTIVE_ZMWS: 25165824,
        LoadingAttributesAliases.P_0: 6257733,
        LoadingAttributesAliases.P_1: 18766925,
        LoadingAttributesAliases.P_2: 141166,
    }
    return ProductivityMetrics.model_validate(data)


@pytest.fixture
def pac_bio_polymerase_metrics() -> PolymeraseMetrics:
    data: dict[str, any] = {
        PolymeraseDataAttributeAliases.MEAN_READ_LENGTH: 85641,
        PolymeraseDataAttributeAliases.READ_LENGTH_N50: 168750,
        PolymeraseDataAttributeAliases.MEAN_LONGEST_SUBREAD_LENGTH: 18042,
        PolymeraseDataAttributeAliases.LONGEST_SUBREAD_LENGTH_N50: 21750,
    }
    return PolymeraseMetrics.model_validate(data)


@pytest.fixture
def pac_bio_smrtlink_databases_metrics(
    barcoded_smrt_cell_internal_id: str,
    pacbio_barcoded_1_c01_cell_full_name: str,
) -> SmrtlinkDatasetsMetrics:
    data: dict[str, any] = {
        SmrtLinkDatabasesAliases.CELL_ID: barcoded_smrt_cell_internal_id,
        SmrtLinkDatabasesAliases.CELL_INDEX: 2,
        SmrtLinkDatabasesAliases.MOVIE_NAME: pacbio_barcoded_1_c01_cell_full_name,
        SmrtLinkDatabasesAliases.PATH: "/srv/cg_data/pacbio/r84202_20240913_121403/1_C01/pb_formats/m84202_240913_162115_s3.hifi_reads.consensusreadset.xml",
        SmrtLinkDatabasesAliases.RUN_COMPLETED_AT: "2024-09-15T14:11:32.418Z",
        SmrtLinkDatabasesAliases.WELL_NAME: "C01",
        SmrtLinkDatabasesAliases.WELL_SAMPLE_NAME: "2023-17834-05",
    }
    return SmrtlinkDatasetsMetrics.model_validate(data)


@pytest.fixture
def pacbio_barcodes_metrics(
    barcoded_smrt_cell_internal_id: str,
    pacbio_barcoded_sample_internal_id: str,
    pacbio_barcoded_1_c01_cell_full_name: str,
) -> BarcodeMetrics:
    data: dict[str, any] = {
        BarcodeMetricsAliases.BARCODED_HIFI_READS: 7785983,
        BarcodeMetricsAliases.BARCODED_HIFI_READS_PERCENTAGE: 0.9962486409672513,
        BarcodeMetricsAliases.BARCODED_HIFI_YIELD: 114676808325,
        BarcodeMetricsAliases.BARCODED_HIFI_YIELD_PERCENTAGE: 0.9963122400514475,
        BarcodeMetricsAliases.BARCODED_MEAN_READ_LENGTH: 14728,
        BarcodeMetricsAliases.UNBARCODED_HIFI_READS: 29318,
        BarcodeMetricsAliases.UNBARCODED_HIFI_YIELD: 424465869,
        BarcodeMetricsAliases.UNBARCODED_HIFI_MEAN_READ_LENGTH: 14477,
    }
    return BarcodeMetrics.model_validate(data)


@pytest.fixture
def pacbio_barcoded_sample_metrics(
    barcoded_smrt_cell_internal_id: str,
    pacbio_barcoded_sample_internal_id: str,
    pacbio_barcoded_1_c01_cell_full_name: str,
) -> SampleMetrics:
    data: dict[str, any] = {
        SampleMetricsAliases.BARCODE_NAME: "bc2004--bc2004",
        SampleMetricsAliases.HIFI_MEAN_READ_LENGTH: 14728,
        SampleMetricsAliases.HIFI_READ_QUALITY: "Q35",
        SampleMetricsAliases.HIFI_READS: 7785983,
        SampleMetricsAliases.HIFI_YIELD: 114676808325,
        SampleMetricsAliases.POLYMERASE_READ_LENGTH: 163451,
        SampleMetricsAliases.SAMPLE_INTERNAL_ID: pacbio_barcoded_sample_internal_id,
    }
    return SampleMetrics.model_validate(data)


@pytest.fixture
def pacbio_unassigned_sample_metrics(
    barcoded_smrt_cell_internal_id: str,
    pacbio_unassigned_sample_internal_id: str,
    pacbio_barcoded_1_c01_cell_full_name: str,
) -> SampleMetrics:
    data: dict[str, any] = {
        SampleMetricsAliases.BARCODE_NAME: "Not Barcoded",
        SampleMetricsAliases.HIFI_MEAN_READ_LENGTH: 14477,
        SampleMetricsAliases.HIFI_READ_QUALITY: "Q28",
        SampleMetricsAliases.HIFI_READS: 29318,
        SampleMetricsAliases.HIFI_YIELD: 424465869,
        SampleMetricsAliases.POLYMERASE_READ_LENGTH: 142228,
        SampleMetricsAliases.SAMPLE_INTERNAL_ID: pacbio_unassigned_sample_internal_id,
    }
    return SampleMetrics.model_validate(data)


@pytest.fixture
def pac_bio_metrics(
    pac_bio_read_metrics: ReadMetrics,
    pac_bio_control_metrics: ControlMetrics,
    pac_bio_productivity_metrics: ProductivityMetrics,
    pac_bio_polymerase_metrics: PolymeraseMetrics,
    pac_bio_smrtlink_databases_metrics: SmrtlinkDatasetsMetrics,
    pacbio_barcodes_metrics: BarcodeMetrics,
    pacbio_barcoded_sample_metrics: SampleMetrics,
    pacbio_unassigned_sample_metrics: SampleMetrics,
) -> PacBioMetrics:
    return PacBioMetrics(
        read=pac_bio_read_metrics,
        control=pac_bio_control_metrics,
        productivity=pac_bio_productivity_metrics,
        polymerase=pac_bio_polymerase_metrics,
        dataset_metrics=pac_bio_smrtlink_databases_metrics,
        barcodes=pacbio_barcodes_metrics,
        samples=[pacbio_barcoded_sample_metrics, pacbio_unassigned_sample_metrics],
    )



from pathlib import Path

import pytest

from cg.io.csv import (
    read_csv,
    read_csv_stream,
    write_csv,
    write_csv_from_dict,
    write_csv_stream,
)
from tests.io.conftest import FileRepresentation


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_get_content_from_file(delimiter: str, delimiter_map: dict[str, FileRepresentation]):
    """
    Tests reading a file with the given separator.
    """
    # GIVEN a file with the given delimiter
    file_path = delimiter_map[delimiter].filepath

    # WHEN reading the file
    raw_csv_content: list[list[str]] = read_csv(file_path=file_path, delimiter=delimiter)

    # THEN assert a list is returned
    assert isinstance(raw_csv_content, list)

    # THEN all three values in each line should be parsed
    assert all(len(line) == 3 for line in raw_csv_content)


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_get_content_from_file_to_dict(
    delimiter: str, delimiter_map: dict[str, FileRepresentation]
):
    """
    Tests reading a delimited file into a list of dictionaries.
    """
    # GIVEN a file with the given delimiter
    file_path = delimiter_map[delimiter].filepath

    # WHEN reading the file
    raw_csv_content: list[list[str]] = read_csv(
        file_path=file_path, delimiter=delimiter, read_to_dict=True
    )

    # THEN assert a list is returned and that the first element is a dict
    assert isinstance(raw_csv_content, list)
    assert isinstance(raw_csv_content[0], dict)

    # THEN all three values in each line should be parsed
    assert all(len(line) == 3 for line in raw_csv_content)


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_get_content_from_stream(delimiter: str, delimiter_map: dict[str, FileRepresentation]):
    """
    Tests reading a delimited stream.
    """
    # GIVEN a string separated by the given delimiter
    stream = delimiter_map[delimiter].content

    # WHEN reading the content in string
    raw_content: list[list[str]] = read_csv_stream(stream=stream, delimiter=delimiter)

    # THEN assert a list is returned
    assert isinstance(raw_content, list)

    # THEN the content should match the expected content
    expected_content = [["Lorem", "ipsum", "sit", "amet"]]
    assert raw_content == expected_content


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_write_csv(delimiter: str, delimiter_map: dict[str, FileRepresentation]):
    """
    Tests writing content to a file with each delimiter.
    """
    # GIVEN a file with the given delimiter

    # GIVEN a file path to write to
    file_path = delimiter_map[delimiter].filepath
    output_file = delimiter_map[delimiter].output_file
    # WHEN reading the file
    raw_csv_content: list[list[str]] = read_csv(file_path=file_path, delimiter=delimiter)

    # WHEN writing the content
    write_csv(content=raw_csv_content, file_path=output_file, delimiter=delimiter)

    # THEN a file was successfully created
    assert Path.exists(output_file)

    # WHEN reading it again
    written_raw_csv_content: list[list[str]] = read_csv(file_path=output_file, delimiter=delimiter)

    # THEN assert that all data is the same
    assert raw_csv_content == written_raw_csv_content


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_write_csv_from_dict(delimiter: str, delimiter_map: dict[str, FileRepresentation]):
    """
    Tests writing content to a file with each delimiter.
    """
    # GIVEN a file with the given delimiter

    # GIVEN a file path to write to
    file_path = delimiter_map[delimiter].filepath
    output_file = delimiter_map[delimiter].output_file

    # WHEN reading the file
    raw_csv_content: list[dict] = read_csv(
        file_path=file_path, delimiter=delimiter, read_to_dict=True
    )

    # WHEN writing the content
    write_csv_from_dict(
        content=raw_csv_content,
        file_path=output_file,
        delimiter=delimiter,
        fieldnames=sorted(raw_csv_content[0].keys()),
    )

    # THEN a file was successfully created
    assert Path.exists(output_file)

    # WHEN reading it again
    written_raw_csv_content: list[dict] = read_csv(
        file_path=output_file, delimiter=delimiter, read_to_dict=True
    )

    # THEN assert that all data is the same
    assert raw_csv_content == written_raw_csv_content


@pytest.mark.parametrize(
    "delimiter",
    [
        ",",
        "\t",
    ],
)
def test_write_csv_stream(delimiter: str, delimiter_map: dict[str, FileRepresentation]):
    """
    Tests writing content to a stream with each delimiter.
    """
    # GIVEN a list of lists
    stream = delimiter_map[delimiter].content
    # WHEN writing the stream
    written_stream: str = write_csv_stream(
        content=[["Lorem", "ipsum", "sit", "amet"]], delimiter=delimiter
    )

    # THEN assert that the stream is correct
    assert written_stream == stream + "\n"



from pathlib import Path

from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_validator.pacbio_run_validator import PacBioRunValidator


def test_pacbio_run_validator(tmp_path: Path, mock_pacbio_run_validator: PacBioRunValidator):
    """Test that the run validator flow works as expected."""
    # GIVEN run data and a run validator
    run_data = PacBioRunData(
        full_path=tmp_path,
        sequencing_run_name="not_relevant",
        well_name="not_relevant",
        plate=1,
    )

    # WHEN ensuring post-processing can start
    mock_pacbio_run_validator.ensure_post_processing_can_start(run_data)

    # THEN the run is validated
    assert mock_pacbio_run_validator.file_manager.get_run_validation_files.call_count == 1
    assert mock_pacbio_run_validator.file_transfer_validator.validate_file_transfer.call_count == 1
    assert mock_pacbio_run_validator.decompressor.decompress.call_count == 1
    assert Path(run_data.full_path, PacBioDirsAndFiles.RUN_IS_VALID).exists()


def test_pacbio_run_validator_skip_if_validated(
    tmp_path: Path, mock_pacbio_run_validator: PacBioRunValidator
):
    """Test that the run validator skips validation if the run is already validated."""
    # GIVEN run data and a run validator
    run_data = PacBioRunData(
        full_path=tmp_path,
        sequencing_run_name="not_relevant",
        well_name="not_relevant",
        plate=1,
    )

    # GIVEN that the run is already validated
    mock_pacbio_run_validator._touch_is_validated(run_data.full_path)

    # WHEN ensuring post-processing can start
    mock_pacbio_run_validator.ensure_post_processing_can_start(run_data)

    # THEN the validation flow is skipped as the run is already validated
    assert Path(run_data.full_path, PacBioDirsAndFiles.RUN_IS_VALID).exists()
    assert mock_pacbio_run_validator.file_manager.get_run_validation_files.call_count == 0
    assert mock_pacbio_run_validator.file_transfer_validator.validate_file_transfer.call_count == 0
    assert mock_pacbio_run_validator.decompressor.decompress.call_count == 0



"""Loqusdb input files fixtures."""

from pathlib import Path

import pytest

from cg.models.observations.input_files import (
    BalsamicObservationsInputFiles,
    MipDNAObservationsInputFiles,
    RarediseaseObservationsInputFiles,
)


@pytest.fixture
def balsamic_observations_input_files_raw(case_id: str, filled_file: Path) -> dict[str, Path]:
    """Return raw observations input files for cancer."""
    return {
        "snv_germline_vcf_path": filled_file,
        "snv_vcf_path": filled_file,
        "sv_germline_vcf_path": filled_file,
        "sv_vcf_path": filled_file,
    }


@pytest.fixture
def balsamic_observations_input_files(
    balsamic_observations_input_files_raw: dict[str, Path],
) -> BalsamicObservationsInputFiles:
    """Return raw observations input files for cancer WHOLE_GENOME_SEQUENCING analysis."""
    return BalsamicObservationsInputFiles(**balsamic_observations_input_files_raw)


@pytest.fixture
def mip_dna_observations_input_files_raw(case_id: str, filled_file: Path) -> dict[str, Path]:
    """Return raw observations input files for rare diseases."""
    return {
        "family_ped_path": filled_file,
        "profile_vcf_path": filled_file,
        "snv_vcf_path": filled_file,
        "sv_vcf_path": None,
    }


@pytest.fixture
def mip_dna_observations_input_files(
    mip_dna_observations_input_files_raw: dict[str, Path],
) -> MipDNAObservationsInputFiles:
    """Return raw observations input files for rare diseases WHOLE_EXOME_SEQUENCING analysis."""
    return MipDNAObservationsInputFiles(**mip_dna_observations_input_files_raw)


@pytest.fixture
def raredisease_observations_input_files_raw(
    case_id: str, filled_file: Path
) -> dict[str, Path | None]:
    """Return raw observations input files for RAREDISEASE."""
    return {
        "family_ped_path": filled_file,
        "profile_vcf_path": filled_file,
        "snv_vcf_path": filled_file,
        "sv_vcf_path": None,
    }


@pytest.fixture
def raredisease_observations_input_files(
    raredisease_observations_input_files_raw: dict[str, Path],
) -> RarediseaseObservationsInputFiles:
    """Return raw observations input files for RAREDISEASE WHOLE_EXOME_SEQUENCING analysis."""
    return RarediseaseObservationsInputFiles(**raredisease_observations_input_files_raw)
